from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from .parsing import *
import math
from ..tradedb import System, Station, TradeDB
from ..tradeexcept import TradeException

######################################################################
# Parser config

help='Calculate a route between two systems.'
name='nav'
epilog=None
wantsTradeDB=True
arguments = [
    ParseArgument('starting', help='System to start from', type=str),
    ParseArgument('ending', help='System to end at', type=str),
]
switches = [
    ParseArgument('--ly-per',
        help='Maximum light years per jump.',
        dest='maxLyPer',
        metavar='N.NN',
        type=float,
    ),
    AvoidPlacesArgument(),
    ParseArgument('--via',
        help='Require specified systems/stations to be en-route (in order).',
        action='append',
        metavar='PLACE[,PLACE,...]',
    ),
    ParseArgument('--stations', '-S',
        help='Include station details.',
        action='store_true',
    ),
    ParseArgument('--refuel-jumps',
        help='Require a station after this many jumps',
        type=int,
        dest='stationInterval',
    ),
    PadSizeArgument(),
    MutuallyExclusiveGroup(
        NoPlanetSwitch(),
        PlanetaryArgument(),
    ),
]

######################################################################
# Helpers


class NoRouteError(TradeException):
    pass


######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    from .commandenv import ResultRow

    srcSystem, dstSystem = cmdenv.origPlace, cmdenv.destPlace
    if isinstance(srcSystem, Station):
        srcSystem = srcSystem.system
    if isinstance(dstSystem, Station):
        dstSystem = dstSystem.system

    maxLyPer = cmdenv.maxLyPer or tdb.maxSystemLinkLy

    cmdenv.DEBUG0("Route from {} to {} with max {}ly per jump.",
                    srcSystem.name(), dstSystem.name(), maxLyPer)

    # Build a list of src->dst pairs
    hops = [ [ srcSystem, None ] ]
    if cmdenv.viaPlaces:
        for hop in cmdenv.viaPlaces:
            hops[-1][1] = hop
            hops.append([hop, None])
    hops[-1][1] = dstSystem

    avoiding = [
        avoid for avoid in cmdenv.avoidPlaces
        if isinstance(avoid, System)
    ]

    route = [ ]
    stationInterval = cmdenv.stationInterval
    for hop in hops:
        hopRoute = tdb.getRoute(
                hop[0], hop[1],
                maxLyPer,
                avoiding,
                stationInterval=stationInterval,
                )
        if not hopRoute:
            raise NoRouteError(
                    "No route found between {} and {} "
                    "with a max {}ly/jump limit.".format(
                        hop[0].name(), hop[1].name(),
                        maxLyPer,
            ))
        route = route[:-1] + hopRoute

    results.summary = ResultRow(
                fromSys=srcSystem,
                toSys=dstSystem,
                maxLy=maxLyPer,
            )

    lastSys, totalLy, dirLy = srcSystem, 0.00, 0.00
    maxPadSize = cmdenv.padSize
    planetary = cmdenv.planetary
    noPlanet = cmdenv.noPlanet

    for (jumpSys, dist) in route:
        jumpLy = lastSys.distanceTo(jumpSys)
        totalLy += jumpLy
        if cmdenv.detail:
            dirLy = jumpSys.distanceTo(dstSystem)
        row = ResultRow(
            action='Via',
            system=jumpSys,
            jumpLy=jumpLy,
            totalLy=totalLy,
            dirLy=dirLy,
            )
        row.stations = []
        if cmdenv.stations:
            for (station) in jumpSys.stations:
                if maxPadSize and not station.checkPadSize(maxPadSize):
                    continue
                if planetary and not station.checkPlanetary(planetary):
                    continue
                if noPlanet and station.planetary != 'N':
                    continue
                rr = ResultRow(
                    station=station,
                    age=station.itemDataAgeStr,
                )
                row.stations.append(rr)
        results.rows.append(row)
        lastSys = jumpSys
    results.rows[0].action='Depart'
    results.rows[-1].action='Arrive'

    return results

######################################################################
# Transform result set into output

def render(results, cmdenv, tdb):
    from ..formatting import RowFormat, ColumnFormat

    if cmdenv.quiet > 1:
        print(','.join(row.system.name() for row in results.rows))
        return

    longestNamed = max(results.rows,
                    key=lambda row: len(row.system.name()))
    longestNameLen = len(longestNamed.system.name())

    rowFmt = RowFormat()
    if cmdenv.detail:
        rowFmt.addColumn("Action", '<', 6, post=":", key=lambda row: row.action)
    rowFmt.addColumn("System", '<', longestNameLen,
            key=lambda row: row.system.name())
    rowFmt.addColumn("JumpLy", '>', '7', '.2f',
            key=lambda row: row.jumpLy)
    if cmdenv.detail:
        rowFmt.addColumn("Stations", '>', 2,
            key=lambda row: len(row.system.stations))
    if cmdenv.detail:
        rowFmt.addColumn("DistLy", '>', '7', '.2f',
            key=lambda row: row.totalLy)
    if cmdenv.detail > 1:
        rowFmt.addColumn("DirLy", '>', 7, '.2f',
            key=lambda row: row.dirLy)

    showStations = cmdenv.stations
    if showStations:
        stnRowFmt = RowFormat(prefix='  /  ').append(
                ColumnFormat("Station", '<', 38,
                    key=lambda row: row.station.dbname)
        ).append(
                ColumnFormat("StnLs", '>', '10',
                    key=lambda row: row.station.distFromStar())
        ).append(
                ColumnFormat("Age/days", '>', 7,
                        key=lambda row: row.age)
        ).append(
                ColumnFormat('Mkt', '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.market])
        ).append(
                ColumnFormat("BMk", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.blackMarket])
        ).append(
                ColumnFormat("Shp", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.shipyard])
        ).append(
                ColumnFormat("Out", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.outfitting])
        ).append(
                ColumnFormat("Arm", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.rearm])
        ).append(
                ColumnFormat("Ref", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.refuel])
        ).append(
                ColumnFormat("Rep", '>', '3',
                    key=lambda row: \
                        TradeDB.marketStates[row.station.repair])
        ).append(
                ColumnFormat("Pad", '>', '3',
                    key=lambda row: \
                        TradeDB.padSizes[row.station.maxPadSize])
        ).append(
                ColumnFormat("Plt", '>', '3',
                    key=lambda row: \
                        TradeDB.planetStates[row.station.planetary])
        )
        if cmdenv.detail > 1:
            stnRowFmt.append(
                ColumnFormat("Itms", ">", 4,
                    key=lambda row: row.station.itemCount)
            )

    if not cmdenv.quiet:
        heading, underline = rowFmt.heading()
        if showStations:
            print(heading)
            heading, underline = stnRowFmt.heading()
        print(heading, underline, sep='\n')

    for row in results.rows:
        print(rowFmt.format(row))
        for stnRow in row.stations:
            print(stnRowFmt.format(stnRow))

    return results



# --------------------------------------------------------------------
# Copyright (C) Oliver 'kfsone' Smith 2014 <oliver@kfs.org>:
# Copyright (C) Bernd 'Gazelle' Gollesch 2016, 2017
# Copyright (C) Jonathan 'eyeonus' Jones 2018
#
# You are free to use, redistribute, or even print and eat a copy of
# this software so long as you include this copyright notice.
# I guarantee there is at least one bug neither of us knew about.
# --------------------------------------------------------------------
# TradeDangerous :: Modules :: Profit Calculator

"""
TradeCalc provides a class for calculating trade loads, hops or
routes, along with some amount of state.

The intent was for it to carry a larger amount of state but
much of that got moved into TradeEnv, so right now TradeCalc
looks a little odd.

Significant Functions:

    Tradecalc.getBestHops
        Finds the best "next hop"s given a set of routes.

Classes:

    TradeCalc
        Encapsulates the calculation functions and item-trades,
    
    Route
        Describes a sequence of trade hops.
    
    TradeLoad
        Describe a cargo load to be carried on a hop.
"""

######################################################################
# Imports

from collections import defaultdict
from collections import namedtuple
from .tradedb import System, Station, Trade, TradeDB, describeAge
from .tradedb import Destination
from .tradeexcept import TradeException

import datetime
import locale
import math
import os
from .misc import progress as pbar
import re
import sys
import time

locale.setlocale(locale.LC_ALL, '')

######################################################################
# Exceptions


class BadTimestampError(TradeException):
    def __init__(
            self,
            tdb,
            stationID, itemID,
            modified
            ):
        self.station = tdb.stationByID[stationID]
        self.item = tdb.itemByID[itemID]
        self.modified = modified
    
    def __str__(self):
        return (
            "Error loading price data from the local db:\n"
            "{} has a StationItem entry for \"{}\" with an invalid "
            "modified timestamp: '{}'.".format(
                self.station.name(),
                self.item.name(),
                str(self.modified),
            )
        )

class NoHopsError(TradeException):
    pass

######################################################################
# Stuff that passes for classes (but isn't)

class TradeLoad(namedtuple('TradeLoad', (
        'items', 'gainCr', 'costCr', 'units'
        ))):
    """
    Describes the manifest of items to be exchanged in a
    trade.
    
    Public attributes:
        items
            A list of (item,qty) tuples tracking the load.
        gainCr
            The predicted total gain in credits
        costCr
            How much this load was bought for
        units
            Total of all the qty values in the items list.
    """
    def __bool__(self):
        return self.units > 0
    
    def __lt__(self, rhs):
        if self.gainCr < rhs.gainCr:
            return True
        if rhs.gainCr < self.gainCr:
            return False
        if self.units < rhs.units:
            return True
        if rhs.units < self.units:
            return False
        return self.costCr < rhs.costCr
    
    @property
    def gpt(self):
        return self.gainCr / self.units if self.units else 0
    
emptyLoad = TradeLoad((), 0, 0, 0)

######################################################################
# Classes

class Route(object):
    """
    Describes a series of hops where a TradeLoad is picked up at
    one station, the player travels via 0 or more hyperspace
    jumps and docks at a second station where they unload.
    E.g. 10 Algae + 5 Hydrogen at Station A, jump to System2,
    jump to System3, dock at Station B, sell everything, buy gold,
    jump to system4 and sell everything at Station X.
    """
    __slots__ = ('route', 'hops', 'startCr', 'gainCr', 'jumps', 'score')
    
    def __init__(self, stations, hops, startCr, gainCr, jumps, score):
        assert stations
        self.route = stations
        self.hops = hops
        self.startCr = startCr
        self.gainCr = gainCr
        self.jumps = jumps
        self.score = score
    
    @property
    def firstStation(self):
        """ Returns the first station in the route. """
        return self.route[0]
    
    @property
    def firstSystem(self):
        """ Returns the first system in the route. """
        return self.route[0].system
    
    @property
    def lastStation(self):
        """ Returns the last station in the route. """
        return self.route[-1]
    
    @property
    def lastSystem(self):
        """ Returns the last system in the route. """
        return self.route[-1].system
    
    @property
    def avggpt(self):
        if self.hops:
            return sum(hop.gpt for hop in self.hops) // len(self.hops)
        return 0
    
    @property
    def gpt(self):
        if self.hops:
            return (
                sum(hop.gainCr for hop in self.hops) //
                sum(hop.units for hop in self.hops)
            )
        return 0
    
    def plus(self, dst, hop, jumps, score):
        """
        Returns a new route describing the sum of this route plus a new hop.
        """
        return Route(
            self.route + (dst,),
            self.hops + (hop,),
            self.startCr,
            self.gainCr + hop[1],
            self.jumps + (jumps,),
            self.score + score,
        )
    
    def __lt__(self, rhs):
        # One route is less than the other if it has a higher score,
        # or the scores are even and the number of jumps are shorter.
        if self.score == rhs.score:
            return len(self.jumps) < len(rhs.jumps)
        return self.score > rhs.score
    
    def __eq__(self, rhs):
        return self.score == rhs.score and len(self.jumps) == len(rhs.jumps)
    
    def str(self):
        return "%s -> %s" % (self.firstStation.name(), self.lastStation.name())
    
    def detail(self, tdenv):
        """
        Return a string describing this route to a given level of detail.
        """
        
        detail, goalSystem = tdenv.detail, tdenv.goalSystem
        
        credits = self.startCr + (tdenv.insurance or 0)
        gainCr = 0
        route = self.route
        
        hops = self.hops
        
        # TODO: Write as a comprehension, just can't wrap my head
        # around it this morning.
        def genSubValues():
            for hop in hops:
                for (tr, qty) in hop[0]:
                    yield len(tr.name(detail))
        longestNameLen = max(genSubValues())
        
        text = self.str()
        if detail >= 1:
            text += " (score: {:f})".format(self.score)
        text += "\n"
        jumpsFmt = ("  Jump {jumps}\n")
        cruiseFmt = ("  Supercruise to {stn}\n")
        distFmt = None
        if detail > 1:
            if detail > 2:
                text += self.summary() + "\n"
                if tdenv.maxJumpsPer > 1:
                    distFmt = (
                        "  Direct: {dist:0.2f}ly, Trip: {trav:0.2f}ly\n"
                    )
            hopFmt = "  Load from {station}:\n{purchases}"
            hopStepFmt = (
                "     {qty:>4} x {item:<{longestName}} "
                "{eacost:>8n}cr vs {easell:>8n}cr, "
                "{age}"
            )
            if detail > 2:
                hopStepFmt += ", total: {ttlcost:>10n}cr"
            hopStepFmt += "\n"
            if not tdenv.summary:
                dockFmt = (
                    "  Unload at {station} => Gain {gain:n}cr "
                    "({tongain:n}cr/ton) => {credits:n}cr\n"
                )
            else:
                jumpsFmt = re.sub("  ", "    ", jumpsFmt, re.M)
                cruiseFmt = re.sub("  ", "    ", cruiseFmt, re.M)
                if distFmt:
                    distFmt = re.sub("  ", "    ", distFmt, re.M)
                hopFmt = "\n" + hopFmt
                dockFmt = "    Expect to gain {gain:n}cr ({tongain:n}cr/ton)\n"
            footer = '  ' + '-' * 76 + "\n"
            endFmt = (
                "Finish at {station} "
                "gaining {gain:n}cr ({tongain:n}cr/ton) "
                "=> est {credits:n}cr total\n"
            )
        elif detail:
            hopFmt = "  Load from {station}:{purchases}\n"
            hopStepFmt = " {qty} x {item} (@{eacost}cr),"
            footer = None
            dockFmt = "  Dock at {station}\n"
            endFmt = (
                "  Finish {station} "
                "+ {gain:n}cr ({tongain:n}cr/ton)"
                "=> {credits:n}cr\n"
            )
        else:
            hopFmt = "  {station}:{purchases}\n"
            hopStepFmt = " {qty} x {item},"
            footer = None
            dockFmt = None
            endFmt = "  {station} +{gain:n}cr ({tongain:n}/ton)"
        
        def jumpList(jumps):
            text, last = "", None
            travelled = 0.
            for jump in jumps:
                if last:
                    dist = last.distanceTo(jump)
                    if dist:
                        if tdenv.detail:
                            text += ", {:.2f}ly -> ".format(dist)
                        else:
                            text += " -> "
                    else:
                        text += " >>> "
                    travelled += dist
                text += jump.name()
                last = jump
            return travelled, text
        
        if detail > 1:
            def decorateStation(station):
                details = []
                if station.lsFromStar:
                    details.append(station.distFromStar(True))
                if station.blackMarket != '?':
                    details.append('BMk:'+station.blackMarket)
                if station.maxPadSize != '?':
                    details.append('Pad:'+station.maxPadSize)
                if station.planetary != '?':
                    details.append('Plt:'+station.planetary)
                if station.shipyard != '?':
                    details.append('Shp:'+station.shipyard)
                if station.outfitting != '?':
                    details.append('Out:'+station.outfitting)
                if station.refuel != '?':
                    details.append('Ref:'+station.refuel)
                details = "{} ({})".format(
                    station.name(),
                    ", ".join(details or ["no details"])
                )
                return details
        else:
            def decorateStation(station):
                return station.name()
        
        if detail and goalSystem:
            def goalDistance(station):
                return " [Distance to {}: {:.2f} ly]\n".format(
                    goalSystem.name(),
                    station.system.distanceTo(goalSystem),
                )
        else:
            def goalDistance(station):
                return ""
        
        for i, hop in enumerate(hops):
            hopGainCr, hopTonnes = hop[1], 0
            purchases = ""
            for (trade, qty) in sorted(
                    hop[0],
                    key=lambda tradeOpt: tradeOpt[1] * tradeOpt[0].gainCr,
                    reverse=True
                    ):
                # Are they within 30 minutes of each other?
                if abs(trade.srcAge - trade.dstAge) <= (30*60):
                    age = max(trade.srcAge, trade.dstAge)
                    age = describeAge(age)
                else:
                    srcAge = describeAge(trade.srcAge)
                    dstAge = describeAge(trade.dstAge)
                    age = "{} vs {}".format(srcAge, dstAge)
                purchases += hopStepFmt.format(
                    qty=qty, item=trade.name(detail),
                    eacost=trade.costCr,
                    easell=trade.costCr + trade.gainCr,
                    ttlcost=trade.costCr*qty,
                    longestName=longestNameLen,
                    age=age,
                )
                hopTonnes += qty
            text += goalDistance(route[i])
            text += hopFmt.format(
                station=decorateStation(route[i]),
                purchases=purchases
            )
            if tdenv.showJumps and jumpsFmt and self.jumps[i]:
                startStn = route[i]
                endStn = route[i+1]
                if startStn.system is not endStn.system:
                    fmt = jumpsFmt
                    travelled, jumps = jumpList(self.jumps[i])
                else:
                    fmt = cruiseFmt
                    travelled, jumps = 0., "{start} >>> {stop}".format(
                        start=startStn.name(), stop=endStn.name()
                    )
                text += fmt.format(
                    jumps=jumps,
                    gain=hopGainCr,
                    tongain=hopGainCr / hopTonnes,
                    credits=credits + gainCr + hopGainCr,
                    stn=route[i+1].dbname
                )
                if travelled and distFmt and len(self.jumps[i]) > 2:
                    text += distFmt.format(
                        dist=startStn.system.distanceTo(endStn.system),
                        trav=travelled,
                    )
            if dockFmt:
                stn = route[i+1]
                stnName = stn.name()
                text += dockFmt.format(
                    station=decorateStation(stn),
                    gain=hopGainCr,
                    tongain=hopGainCr / hopTonnes,
                    credits=credits + gainCr + hopGainCr
                )
            
            gainCr += hopGainCr
        
        lastStation = self.lastStation
        if lastStation.system is not goalSystem:
            text += goalDistance(lastStation)
        text += footer or ""
        text += endFmt.format(
            station=decorateStation(lastStation),
            gain=gainCr,
            credits=credits + gainCr,
            tongain=self.gpt
        )
        
        return text
    
    def summary(self):
        """
        Returns a string giving a short summary of this route.
        """
        
        credits, hops, jumps = self.startCr, self.hops, self.jumps
        ttlGainCr = sum(hop[1] for hop in hops)
        numJumps = sum(
            len(hopJumps)-1
            for hopJumps in jumps
            if hopJumps     # don't include in-system hops
        )
        return (
            "Start CR: {start:10n}\n"
            "Hops    : {hops:10n}\n"
            "Jumps   : {jumps:10n}\n"
            "Gain CR : {gain:10n}\n"
            "Gain/Hop: {hopgain:10n}\n"
            "Final CR: {final:10n}\n" . format(
                start=credits,
                hops=len(hops),
                jumps=numJumps,
                gain=ttlGainCr,
                hopgain=ttlGainCr // len(hops),
                final=credits + ttlGainCr
            )
        )

class TradeCalc(object):
    """
    Container for accessing trade calculations with common properties.
    """
    
    def __init__(self, tdb, tdenv=None, fit=None, items=None):
        """
        Constructs the TradeCalc object and loads sell/buy data.
        
        Parameters:
            tdb
                The TradeDB() object to use to access data,
            tdenv [optional]
                TradeEnv() that controls behavior,
            fit [optional]
                Lets you specify a fitting function,
            items [optional]
                Iterable [itemID or Item()] that restricts loading,
        
        TradeEnv options:
            tdenv.avoidItems
                Iterable of [Item] that prevents items being loaded
            tdenv.maxAge
                Maximum age in days of data that gets loaded
            tdenv.supply
                Require at least this much supply to load an item
            tdenv.demand
                Require at least this much demand to load an item
        """
        if not tdenv:
            tdenv = tdb.tdenv
        self.tdb = tdb
        self.tdenv = tdenv
        self.defaultFit = fit or self.simpleFit
        if "BRUTE_FIT" in os.environ:
            self.defaultFit = self.bruteForceFit
        minSupply = self.tdenv.supply or 0
        minDemand = self.tdenv.demand or 0
        
        db = tdb.getDB()
        
        wheres, binds = [], []
        if tdenv.maxAge:
            maxDays = datetime.timedelta(days=tdenv.maxAge)
            cutoff = datetime.datetime.now() - maxDays
            wheres.append("(modified >= ?)")
            binds.append(str(cutoff.replace(microsecond=0)))
        
        if tdenv.avoidItems or items:
            avoidItemIDs = set(item.ID for item in tdenv.avoidItems)
            loadItems = items or tdb.itemByID.values()
            loadItemIDs = set()
            for item in loadItems:
                ID = item if isinstance(item, int) else item.ID
                if ID not in avoidItemIDs:
                    loadItemIDs.add(str(ID))
            if not loadItemIDs:
                raise TradeException("No items to load.")
            loadItemIDs = ",".join(str(ID) for ID in loadItemIDs)
            wheres.append("(item_id IN ({}))".format(loadItemIDs))
        
        demand = self.stationsBuying = defaultdict(list)
        supply = self.stationsSelling = defaultdict(list)
        
        whereClause = " AND ".join(wheres) or "1"
        
        lastStnID, stnAppend = 0, None
        dmdCount, supCount = 0, 0
        stmt = """
                SELECT  station_id, item_id,
                        strftime('%s', modified),
                        demand_price, demand_units, demand_level,
                        supply_price, supply_units, supply_level
                  FROM  StationItem
                 WHERE  {where}
        """.format(where=whereClause)
        tdenv.DEBUG1("TradeCalc loading StationItem values")
        tdenv.DEBUG2("sql: {}, binds: {}", stmt, binds)
        cur = db.execute(stmt, binds)
        now = int(time.time())
        for (stnID, itmID,
                timestamp,
                dmdCr, dmdUnits, dmdLevel,
                supCr, supUnits, supLevel) in cur:
            if stnID != lastStnID:
                dmdAppend = demand[stnID].append
                supAppend = supply[stnID].append
                lastStnID = stnID
            try:
                ageS = now - int(timestamp)
            except TypeError:
                raise BadTimestampError(
                    self.tdb,
                    stnID, itmID, timestamp
                )
            if dmdCr > 0:
                if not minDemand or dmdUnits >= minDemand:
                    dmdAppend((itmID, dmdCr, dmdUnits, dmdLevel, ageS))
                    dmdCount += 1
            if supCr > 0 and supUnits:
                if not minSupply or supUnits >= minSupply:
                    supAppend((itmID, supCr, supUnits, supLevel, ageS))
                    supCount += 1
                    
        tdenv.DEBUG0("Loaded {} buys, {} sells".format(dmdCount, supCount))
    
    def bruteForceFit(self, items, credits, capacity, maxUnits):
        """
        Brute-force generation of all possible combinations of items.
        This is provided to make it easy to validate the results of future
        variants or optimizations of the fit algorithm.
        """
        def _fitCombos(offset, cr, cap, level=1):
            if cr <= 0 or cap <= 0:
                return emptyLoad
            while True:
                if offset >= len(items):
                    return emptyLoad
                item = items[offset]
                offset += 1
                
                itemCost = item.costCr
                maxQty = min(maxUnits, cap, cr // itemCost)
                
                if item.supply < maxQty and item.supply > 0:  # -1 = unknown
                    maxQty = min(maxQty, item.supply)
                
                if maxQty > 0:
                    break
            
            # find items that don't include us
            bestLoad = _fitCombos(offset, cr, cap, level + 1)
            itemGain = item.gainCr
            
            for qty in range(1, maxQty + 1):
                loadGain, loadCost = itemGain * qty, itemCost * qty
                load = TradeLoad(((item, qty),), loadGain, loadCost, qty)
                subLoad = _fitCombos(
                    offset, cr - loadCost, cap - qty, level+1
                )
                combGain = loadGain + subLoad.gainCr
                if combGain < bestLoad.gainCr:
                    continue
                combCost = loadCost + subLoad.costCr
                combUnits = qty + subLoad.units
                if combGain == bestLoad.gainCr:
                    if combUnits > bestLoad.units:
                        continue
                    if combUnits == bestLoad.units:
                        if combCost >= bestLoad.costCr:
                            continue
                bestLoad = TradeLoad(
                    load.items + subLoad.items,
                    combGain, combCost, combUnits
                )
            
            return bestLoad
        
        return _fitCombos(0, credits, capacity)
    
    def fastFit(self, items, credits, capacity, maxUnits):
        """
            Best load calculator using a recursive knapsack-like
            algorithm to find multiple loads and return the best.
            [eyeonus] Left in for the masochists, as this becomes
            horribly slow at stations with many items for sale.
            As in iooks-like-the-program-has-frozen slow.
        """
        
        def _fitCombos(offset, cr, cap):
            """
                Starting from offset, consider a scenario where we
                would purchase the maximum number of each item
                given the cr+cap limitations. Then, assuming that
                load, solve for the remaining cr+cap from the next
                value of offset.
                
                The "best fit" is not always the most profitable,
                so we yield all the results and leave the caller
                to determine which is actually most profitable.
            """
            
            bestGainCr = -1
            bestItem = None
            bestQty = 0
            bestCostCr = 0
            bestSub = None
            
            qtyCeil = min(maxUnits, cap)
            
            for iNo in range(offset, len(items)):
                item = items[iNo]
                itemCostCr = item.costCr
                maxQty = min(qtyCeil, cr // itemCostCr)
                
                if maxQty <= 0:
                    continue
                
                supply = item.supply
                if supply <= 0:
                    continue
                
                maxQty = min(maxQty, supply)
                
                itemGainCr = item.gainCr
                if maxQty == cap:
                    # full load
                    gain = itemGainCr * maxQty
                    if gain > bestGainCr:
                        cost = itemCostCr * maxQty
                        # list is sorted by gain DESC, cost ASC
                        bestGainCr = gain
                        bestItem = item
                        bestQty = maxQty
                        bestCostCr = cost
                        bestSub = None
                    break
                
                loadCostCr = maxQty * itemCostCr
                loadGainCr = maxQty * itemGainCr
                if loadGainCr > bestGainCr:
                    bestGainCr = loadGainCr
                    bestCostCr = loadCostCr
                    bestItem = item
                    bestQty = maxQty
                    bestSub = None
                
                crLeft, capLeft = cr - loadCostCr, cap - maxQty
                if crLeft > 0 and capLeft > 0:
                    # Solve for the remaining credits and capacity with what
                    # is left in items after the item we just checked.
                    subLoad = _fitCombos(iNo+1, crLeft, capLeft)
                    if subLoad is emptyLoad:
                        continue
                    ttlGain = loadGainCr + subLoad.gainCr
                    if ttlGain < bestGainCr:
                        continue
                    ttlCost = loadCostCr + subLoad.costCr
                    if ttlGain == bestGainCr and ttlCost >= bestCostCr:
                        continue
                    bestGainCr = ttlGain
                    bestItem = item
                    bestQty = maxQty
                    bestCostCr = ttlCost
                    bestSub = subLoad
            
            if not bestItem:
                return emptyLoad
            
            bestLoad = ((bestItem, bestQty),)
            if bestSub:
                bestLoad = bestLoad + bestSub.items
                bestQty += bestSub.units
            return TradeLoad(bestLoad, bestGainCr, bestCostCr, bestQty)
        
        return _fitCombos(0, credits, capacity)
    
    # Mark's test run, to spare searching back through the forum posts for it.
    # python trade.py run --fr="Orang/Bessel Gateway" --cap=720 --cr=11b --ly=24.73 --empty=37.61 --pad=L --hops=2 --jum=3 --loop --summary -vv --progress
    def simpleFit(self, items, credits, capacity, maxUnits):
        """
        Simplistic load calculator:
        (The item list is sorted with highest profit margin items in front.)
        Step 1: Fill hold with as much of item1 as possible based on the limiting
                factors of hold size, supply amount, and available credits.
        
        Step 2: If there is space in the hold and money available, repeat Step 1
                with item2, item3, etc. until either the hold is filled
                or the commander is too poor to buy more.
        
        When amount of credits isn't a limiting factor, this should produce
        the most profitable route ~99.7% of the time, and still be very
        close to the most profitable the rest of the time.
        (Very close = not enough less profit that anyone should care,
        especially since this thing doesn't suffer slowdowns like fastFit.)
        """
        
        n = 0
        load = ()
        gainCr = 0
        costCr = 0
        qty = 0
        while n < len(items) and credits > 0 and capacity > 0:
            qtyCeil = min(maxUnits, capacity)
            
            item = items[n]
            maxQty = min(qtyCeil, credits // item.costCr)
            
            if maxQty > 0 and item.supply > 0:
                maxQty = min(maxQty, item.supply)
                
                loadCostCr = maxQty * item.costCr
                loadGainCr = maxQty * item.gainCr
                
                load = load + ((item, maxQty),)
                qty += maxQty
                capacity -= maxQty
                
                gainCr += loadGainCr
                costCr += loadCostCr
                credits -= loadCostCr
            
            n += 1
        
        return TradeLoad(load, gainCr, costCr, qty)
    
    def getTrades(self, srcStation, dstStation, srcSelling=None):
        """
        Returns the most profitable trading options from
        one station to another (uni-directional).
        """
        if not srcSelling:
            srcSelling = self.stationsSelling.get(srcStation.ID, None)
            if not srcSelling:
                return None
        dstBuying = self.stationsBuying.get(dstStation.ID, None)
        if not dstBuying:
            return None
        
        trading = []
        itemIdx = self.tdb.itemByID
        minGainCr = max(1, self.tdenv.minGainPerTon or 1)
        maxGainCr = max(minGainCr, self.tdenv.maxGainPerTon or sys.maxsize)
        getBuy = {buy[0]: buy for buy in dstBuying}.get
        addTrade = trading.append
        for sell in srcSelling: # should be the smaller list
            buy = getBuy(sell[0], None)
            if buy:
                gainCr = buy[1] - sell[1]
                if gainCr >= minGainCr and gainCr <= maxGainCr:
                    addTrade(Trade(
                        itemIdx[sell[0]],
                        sell[1], gainCr,
                        sell[2], sell[3],
                        buy[2], buy[3],
                        sell[4], buy[4],
                    ))
        
        # SORT BY profit DESC, cost ASC
        # So if two items have the same profit, the cheapest will come first.
        trading.sort(key=lambda trade: trade.costCr)
        trading.sort(key=lambda trade: trade.gainCr, reverse=True)
        
        return trading
    
    def getBestHops(self, routes, restrictTo=None):
        """
        Given a list of routes, try all available next hops from each
        route.
        
        Store the results by destination so that we pick the
        best route-to-point for each destination at each step.
        
        If we have two routes: A->B->D, A->C->D and A->B->D produces
        more profit, there's no point continuing the A->C->D path.
        """
        
        tdb = self.tdb
        tdenv = self.tdenv
        avoidPlaces = getattr(tdenv, 'avoidPlaces', None) or ()
        assert not restrictTo or isinstance(restrictTo, set)
        maxJumpsPer = tdenv.maxJumpsPer
        maxLyPer = tdenv.maxLyPer
        maxPadSize = tdenv.padSize
        planetary = tdenv.planetary
        noPlanet = tdenv.noPlanet
        maxLsFromStar = tdenv.maxLs or float('inf')
        reqBlackMarket = getattr(tdenv, 'blackMarket', False) or False
        maxAge = getattr(tdenv, 'maxAge') or 0
        credits = tdenv.credits - (getattr(tdenv, 'insurance', 0) or 0)
        fitFunction = self.defaultFit
        capacity = tdenv.capacity
        maxUnits = getattr(tdenv, 'limit') or capacity
        
        bestToDest = {}
        safetyMargin = 1.0 - tdenv.margin
        unique = tdenv.unique
        loopInt = getattr(tdenv, 'loopInt', 0) or None
        
        # Penalty is expressed as percentage, reduce it to a multiplier
        if tdenv.lsPenalty:
            lsPenalty = max(min(tdenv.lsPenalty / 100, 1), 0)
        else:
            lsPenalty = 0
        
        goalSystem = tdenv.goalSystem
        uniquePath = None
        
        restrictStations = set()
        if restrictTo:
            for place in restrictTo:
                if isinstance(place, Station):
                    restrictStations.add(place)
                elif isinstance(place, System) and place.stations:
                    restrictStations.update(place.stations)
        
        # Are we doing direct routes?
        if tdenv.direct:
            if goalSystem and not restrictTo:
                restrictTo = (goalSystem,)
                restrictStations = set(goalSystem.stations)
            if avoidPlaces:
                restrictStations = set(
                    stn for stn in restrictStations
                    if stn not in avoidPlaces and \
                        stn.system not in avoidPlaces
                )
            def station_iterator(srcStation):
                srcSys = srcStation.system
                srcDist = srcSys.distanceTo
                for stn in restrictStations:
                    stnSys = stn.system
                    yield Destination(
                        stnSys, stn,
                        (srcSys, stnSys),
                        srcDist(stnSys)
                    )
        else:
            getDestinations = tdb.getDestinations
            def station_iterator(srcStation):
                yield from getDestinations(
                    srcStation,
                    maxJumps=maxJumpsPer,
                    maxLyPer=maxLyPer,
                    avoidPlaces=avoidPlaces,
                    maxPadSize=maxPadSize,
                    maxLsFromStar=maxLsFromStar,
                    noPlanet=noPlanet,
                    planetary=planetary,
                )
        
        prog = pbar.Progress(len(routes), 25)
        connections = 0
        getSelling = self.stationsSelling.get
        for route in routes:
            if tdenv.progress:
                prog.increment(1)
            tdenv.DEBUG1("Route = {}", route.str())
            
            srcStation = route.lastStation
            startCr = credits + int(route.gainCr * safetyMargin)
            routeJumps = len(route.jumps)
            
            srcSelling = getSelling(srcStation.ID, None)
            srcSelling = tuple(
                values for values in srcSelling
                if values[1] <= startCr
            )
            if not srcSelling:
                tdenv.DEBUG1("Nothing sold/affordable - next.")
                continue
            
            if goalSystem:
                origSystem = route.firstSystem
                srcSystem = srcStation.system
                srcDistTo = srcSystem.distanceTo
                goalDistTo = goalSystem.distanceTo
                origDistTo = origSystem.distanceTo
                srcGoalDist = srcDistTo(goalSystem)
                srcOrigDist = srcDistTo(origSystem)
                origGoalDist = origDistTo(goalSystem)
            
            if unique:
                uniquePath = route.route
            elif loopInt:
                uniquePath = route.route[-loopInt:-1]
            
            stations = (
                dest for dest in station_iterator(srcStation)
                if dest.station != srcStation
            )
            if reqBlackMarket:
                stations = (d for d in stations if d.station.blackMarket == 'Y')
            if uniquePath:
                stations = (d for d in stations if d.station not in uniquePath)
            if restrictStations:
                stations = (
                    d for d in stations
                    if d.station in restrictStations
                )
            if maxAge:
                stations = (d for d in stations if d.station.dataAge)
                stations = (d for d in stations if d.station.dataAge <= maxAge)
            if goalSystem:
                if bool(tdenv.unique):
                    stations = (
                        d for d in stations if d.system is not srcSystem
                    )
                stations = (
                    d for d in stations
                    if d.system is goalSystem or d.distLy < srcGoalDist
                )
            
            if tdenv.debug >= 1:
                def annotate(dest):
                    tdenv.DEBUG1(
                        "destSys {}, destStn {}, jumps {}, distLy {}",
                        dest.system.dbname,
                        dest.station.dbname,
                        "->".join(jump.str() for jump in dest.via),
                        dest.distLy
                    )
                    return True
                stations = (d for d in stations if annotate(d))
            
            for dest in stations:
                dstStation = dest.station
                
                connections += 1
                items = self.getTrades(srcStation, dstStation, srcSelling)
                if not items:
                    continue
                trade = fitFunction(items, startCr, capacity, maxUnits)
                
                multiplier = 1.0
                # Calculate total K-lightseconds supercruise time.
                # This will amortize for the start/end stations
                dstSys = dest.system
                if goalSystem and dstSys is not goalSystem:
                    dstGoalDist = goalDistTo(dstSys)
                    # Biggest reward for shortening distance to goal
                    score = 5000 * origGoalDist / dstGoalDist
                    # bias towards bigger reductions
                    score += 50 * srcGoalDist / dstGoalDist
                    # discourage moving back towards origin
                    if dstSys is not origSystem:
                        score += 10 * (origDistTo(dstSys) - srcOrigDist)
                    # Gain per unit pays a small part
                    score += (trade.gainCr / trade.units) / 25
                else:
                    score = trade.gainCr
                if lsPenalty:
                    # [kfsone] Only want 1dp
                    
                    cruiseKls = int(dstStation.lsFromStar / 100) / 10
                    # Produce a curve that favors distances under 1kls
                    # positively, starts to penalize distances over 1k,
                    # and after 4kls starts to penalize aggressively
                    # http://goo.gl/Otj2XP
                    
                    # [eyeonus] As aadler pointed out, this goes into negative
                    # numbers, which causes problems.
                    #penalty = ((cruiseKls ** 2) - cruiseKls) / 3
                    #penalty *= lsPenalty
                    #multiplier *= (1 - penalty)
                    
                    # [eyeonus]:
                    # (Keep in mind all this ignores values of x<0.)
                    # The sigmoid: (1-(25(x-1))/(1+abs(25(x-1))))/4
                    # ranges between 0.5 and 0 with a drop around x=1,
                    # which makes it great for giving a boost to distances < 1Kls.
                    #
                    # The sigmoid: (-1-(50(x-4))/(1+abs(50(x-4))))/4
                    # ranges between 0 and -0.5 with a drop around x=4,
                    # making it great for penalizing distances > 4Kls.
                    #
                    # The curve: (-1+1/(x+1)^((x+1)/4))/2
                    # ranges between 0 and -0.5 in a smooth arc,
                    # which will be used for making distances
                    # closer to 4Kls get a slightly higher penalty
                    # then distances closer to 1Kls.
                    #
                    # Adding the three together creates a doubly-kinked curve
                    # that ranges from ~0.5 to -1.0, with drops around x=1 and x=4,
                    # which closely matches ksfone's intention without going into
                    # negative numbers and causing problems when we add it to
                    # the multiplier variable. ( 1 + -1 = 0 )
                    #
                    # You can see a graph of the formula here:
                    # https://goo.gl/sn1PqQ
                    # NOTE: The black curve is at a penalty of 0%,
                    # the red curve at a penalty of 100%, with intermediates at
                    # 25%, 50%, and 75%.
                    # The other colored lines show the penalty curves individually
                    # and the teal composite of all three.
                    
                    def sigmoid(x):
                        return x/(1+abs(x))
                    
                    boost = (1 - sigmoid(25 * (cruiseKls - 1))) / 4
                    drop = (-1 - sigmoid(50 * (cruiseKls - 4))) / 4
                    try:
                        penalty = (-1 + 1 / (cruiseKls + 1) ** ((cruiseKls + 1) / 4)) / 2
                    except OverflowError:
                        penalty = -0.5
                    
                    multiplier += (penalty + boost + drop) * lsPenalty
                
                score *= multiplier
                
                dstID = dstStation.ID
                try:
                    # See if there is already a candidate for this destination
                    btd = bestToDest[dstID]
                except KeyError:
                    # No existing candidate, we win by default
                    pass
                else:
                    bestRoute = btd[1]
                    bestScore = btd[5]
                    # Check if it is a better option than we just produced
                    bestTradeScore = bestRoute.score + bestScore
                    newTradeScore = route.score + score
                    if bestTradeScore > newTradeScore:
                        continue
                    if bestTradeScore == newTradeScore:
                        bestLy = btd[4]
                        if bestLy <= dest.distLy:
                            continue
                
                bestToDest[dstID] = (
                    dstStation, route, trade, dest.via, dest.distLy, score
                )
        
        prog.clear()
        
        if connections == 0:
            raise NoHopsError(
                "No destinations could be reached within the constraints."
            )
        
        result = []
        for (dst, route, trade, jumps, ly, score) in bestToDest.values():
            result.append(route.plus(dst, trade, jumps, score))
        
        return result



# Copyright (C) Oliver 'kfsone' Smith 2014 <oliver@kfs.org>:
#  You are free to use, redistribute, or even print and eat a copy of
#  this software so long as you include this copyright notice.
#  I guarantee there is at least one bug neither of us knew about.
#---------------------------------------------------------------------
# TradeDangerous :: Modules :: Multi-function display wrapper
#
#   Multi-Function Display wrappers

from __future__ import absolute_import, with_statement, print_function, division, unicode_literals

######################################################################
# imports

import sys
import time


######################################################################
# exceptions

class MissingDeviceError(Exception):
    """
        Throw when no instance of a device cannot be found.
    """
    pass

######################################################################
# classes

class DummyMFD(object):
    """
        Base class for the MFD drivers, implemented as no-ops so that
        you can always use all MFD functions without conditionals.
    """

    def __init__(self):
        pass


    def finish(self):
        """
            Close down the driver.
        """
        pass


    def display(self, line1, line2="", line3="", delay=None):
        """
            Display data to the MFD.
            Arguments: 1-3 lines of text plus optional pause in seconds.
        """
        pass


    def attention(self, duration):
        """
            Draw the user's attention.
        """
        print("\a")


# for now, I'm going to put the wrapper classes just here. till I
# have a few more to play with and can figure out how I want to
# organize them.

class X52ProMFD(DummyMFD):
    """
        Wrapper for the Saitek X52 Pro MFD.
    """

    def __init__(self):
        from sys import exit
        from . saitek import directoutput, x52pro
        try:
            self.doObj = x52pro.X52Pro()
        except MissingDeviceError:
            print("{}: error: Could not find any X52 Pro devices attached to the system - please ensure your device is connected and drivers are installed.".format(__name__))
            exit(1)
        except directoutput.DLLError as e:
            print("{}: error#{}: Unable to initialize the Saitek X52 Pro module: {}".format(__name__, e.error_code, e.msg), file=sys.stderr)
            exit(1)

        self.page = self.doObj.add_page('TD')
        self.display('TradeDangerous', 'INITIALIZING')


    def finish(self):
        self.doObj.finish()


    def display(self, line1, line2="", line3="", delay=None):
        self.page[0], self.page[1], self.page[2] = line1, line2, line3
        if delay:
            time.sleep(delay)

    def attention(self, duration):
        page = self.page
        iterNo = 0
        cutoff = time.time() + duration
        while time.time() <= cutoff:
            for ledNo in range(0, 20):
                page.set_led(ledNo, (iterNo + ledNo) % 4)
            iterNo += 1
            time.sleep(0.02)



import codecs
import csv
import datetime
import json
import os
import platform
import sqlite3
import time

from urllib import request
from calendar import timegm
from pathlib import Path
from importlib import reload
from builtins import str

from .. import plugins, cache, csvexport, tradedb, tradeenv, transfers
from ..misc import progress as pbar
from ..plugins import PluginException
# Constants

BASE_URL = "http://elite.ripz.org/files/"
FALLBACK_URL = "https://eddb.io/archive/v6/"
SHIPS_URL = "https://raw.githubusercontent.com/EDCD/coriolis-data/master/dist/index.json"
COMMODITIES = "commodities.json"
SYSTEMS = "systems_populated.jsonl"
STATIONS = "stations.jsonl"
UPGRADES = "modules.json"
LISTINGS = "listings.csv"
LIVE_LISTINGS = "listings-live.csv"

class DecodingError(PluginException):
    pass


class ImportPlugin(plugins.ImportPluginBase):
    """
    Plugin that downloads data from eddb.
    """

    pluginOptions = {
        'item':         "Regenerate Categories and Items using latest commodities.json dump.",
        'system':       "Regenerate Systems using latest system-populated.jsonl dump.",
        'station':      "Regenerate Stations using latest stations.jsonl dump. (Implies '-O system')",
        'ship':         "Regenerate Ships using latest coriolis.io json dump.",
        'shipvend':     "Regenerate ShipVendors using latest stations.jsonl dump. (Implies '-O system,station,ship')",
        'upgrade':      "Regenerate Upgrades using latest modules.json dump.",
        'upvend':       "Regenerate UpgradeVendors using latest stations.jsonl dump. (Implies '-O system,station,upgrade')",
        'listings':     "Update market data using latest listings.csv dump. (Implies '-O item,system,station')",
        'all':          "Update everything with latest dumpfiles. (Regenerates all tables)",
        'clean':        "Erase entire database and rebuild from empty. (Regenerates all tables.)",
        'skipvend':     "Don't regenerate ShipVendors or UpgradeVendors. (Supercedes '-O all', '-O clean'.)",
        'force':        "Force regeneration of selected items even if source file not updated since previous run. "
                        "(Useful for updating Vendor tables if they were skipped during a '-O clean' run.)",
        'fallback':     "Fallback to using EDDB.io if Tromador's mirror isn't working.",
        'progbar':      "Does nothing, only included for backwards compatibility.",
        'solo':         "Don't download crowd-sourced market data. (Implies '-O skipvend', supercedes '-O all', '-O clean', '-O listings'.)"
    }

    def __init__(self, tdb, tdenv):
        super().__init__(tdb, tdenv)

        self.dataPath = Path(os.environ.get('TD_EDDB')) or tdb.dataPath / Path("eddb")
        self.commoditiesPath = Path(COMMODITIES)
        self.systemsPath = Path(SYSTEMS)
        self.stationsPath = Path(STATIONS)
        self.upgradesPath = Path(UPGRADES)
        self.listingsPath = Path(LISTINGS)
        self.liveListingsPath = Path(LIVE_LISTINGS)
        self.shipsPath = Path("index.json")
        self.pricesPath = Path("listings.prices")
        self.updated = {
                "Category": False,
                "Item": False,
                "RareItem": False,
                "Ship": False,
                "ShipVendor": False,
                "Station": False,
                "System": False,
                "Upgrade": False,
                "UpgradeVendor": False,
                "Listings": False
            }
    
    def execute(self, sql_cmd, args = None):
        tdb, tdenv = self.tdb, self.tdenv
        cur = tdb.getDB().cursor()
        
        success = False
        result = None
        while not success:
            try:
                if args:
                    result = cur.execute(sql_cmd, args)
                else:
                    result = cur.execute(sql_cmd)
                success = True
            except sqlite3.OperationalError as e:
                if "locked" not in str(e):
                    success = True
                    raise sqlite3.OperationalError(e)
                else:
                    print("(execute) Database is locked, waiting for access.", end = "\r")
                    time.sleep(1)
        return result
    
    def executemany(self, sql_cmd, args):
        tdb, tdenv = self.tdb, self.tdenv
        cur = tdb.getDB().cursor()
        
        success = False
        result = None
        while not success:
            try:
                result = cur.executemany(sql_cmd, args)
                success = True
            except sqlite3.OperationalError as e:
                if "locked" not in str(e):
                    success = True
                    raise sqlite3.OperationalError(e)
                else:
                    print("(execute) Database is locked, waiting for access.", end = "\r")
                    time.sleep(1)
        return result
    
    def fetchIter(self, cursor, arraysize=1000):
        """
        An iterator that uses fetchmany to keep memory usage down
        and speed up the time to retrieve the results dramatically.
        """
        while True:
            results = cursor.fetchmany(arraysize)
            if not results:
                break
            for result in results:
                yield result
            
    def downloadFile(self, urlTail, path):
        """
        Fetch the latest dumpfile from the website if newer than local copy.
        """
        tdb, tdenv = self.tdb, self.tdenv
        
        tdenv.DEBUG0("Checking for update to '{}'.", path)
        if urlTail == SHIPS_URL:
            url = SHIPS_URL
        else:
            if not self.getOption('fallback'):
                try:
                    url = BASE_URL + urlTail
                    response = request.urlopen(url)
                except:
                    # If Tromador's server fails for whatever reason,
                    # fallback to download direct from EDDB.io
                    self.options["fallback"] = True
            if self.getOption('fallback'):
                # EDDB.io doesn't have live listings.
                if urlTail == LIVE_LISTINGS:
                    return False

                url = FALLBACK_URL + urlTail
                response = request.urlopen(url)
        
        dumpModded = 0
        # The coriolis file is from github, so it doesn't have a "Last-Modified" metadata.
        if url != SHIPS_URL:
            # Stupid strptime() is locale-dependent, so we do it ourself.
            # 'Last-Modified' is in the form "DDD, dd MMM yyyy hh:mm:ss GMT"
            # dDL                               0   1   2    3        4   5
            # dTL                                               0  1  2

            # We'll need to turn the 'MMM' into a number.
            Months = {'Jan':1, 'Feb':2, 'Mar':3, 'Apr':4, 'May':5, 'Jun':6, 'Jul':7, 'Aug':8, 'Sep':9, 'Oct':10, 'Nov':11, 'Dec':12}

            # We need to split the string twice, because the time is separated by ':', not ' '.
            dDL = response.getheader("Last-Modified").split(' ')
            dTL = dDL[4].split(':')

            # Now we need to make a datetime object using the DateList and TimeList we just created,
            # and then we can finally convert that to a Unix-epoch number.
            dumpDT = datetime.datetime(int(dDL[3]), Months[dDL[2]], int(dDL[1]),\
               hour=int(dTL[0]), minute=int(dTL[1]), second=int(dTL[2]),\
               tzinfo=datetime.timezone.utc)
            dumpModded = timegm(dumpDT.timetuple())

        if Path.exists(self.dataPath / path):
            localModded = (self.dataPath / path).stat().st_mtime
            if localModded >= dumpModded and url != SHIPS_URL:
                tdenv.DEBUG0("'{}': Dump is not more recent than Local.", path)
                return False
        
        tdenv.NOTE("Downloading file '{}'.", path)
        transfers.download( self.tdenv, url, self.dataPath / path)
        return True

    def importUpgrades(self):
        """
        Populate the Upgrade table using modules.json
        Writes directly to database.
        """
        tdb, tdenv = self.tdb, self.tdenv
        
        tdenv.NOTE("Processing Upgrades: Start time = {}", datetime.datetime.now())
        with open(str(self.dataPath / self.upgradesPath), "rU") as fh:
            upgrades = json.load(fh)
        for upgrade in iter(upgrades):
            upgrade_id = upgrade['id']
            name = upgrade['name'] if upgrade['name'] else upgrade['ed_symbol'].replace('_',' ')
            weight = upgrade['mass'] if 'mass' in upgrade else 0
            cost = upgrade['price'] if upgrade['price'] else 0
            
            tdenv.DEBUG1("Updating: {}, {}, {}, {}", upgrade_id, name, weight, cost)
            try:
                self.execute("""INSERT INTO Upgrade
                            ( upgrade_id,name,weight,cost ) VALUES
                            ( ?, ?, ?, ? ) """,
                            (upgrade_id, name, weight, cost))
            except sqlite3.IntegrityError:
                try:
                    self.execute("""UPDATE Upgrade
                                SET name = ?,weight = ?,cost = ?
                                WHERE upgrade_id = ?""", 
                                (name, weight, cost,
                                 upgrade_id))
                except sqlite3.IntegrityError:
                    tdenv.DEBUG0("Unable to insert or update: {}, {}, {}, {}", upgrade_id, name, weight, cost)

        self.updated['Upgrade'] = True
        
        tdenv.NOTE("Finished processing Upgrades. End time = {}", datetime.datetime.now())

    def importShips(self):
        """
        Populate the Ship table using coriolis.io's index.json
        Writes directly to database.
        """
        tdb, tdenv = self.tdb, self.tdenv
        
        tdenv.NOTE("Processing Ships: Start time = {}", datetime.datetime.now())
        with open(str(self.dataPath / self.shipsPath), "rU") as fh:
            ships = json.load(fh)['Ships']
        for ship in iter(ships):
            ship_id = ships[ship]['eddbID']
            name = ships[ship]['properties']['name']
            cost = ships[ship]['retailCost']
            fdev_id = ships[ship]['edID']
            # Arg. Why you do this to me, EDCD?
            if "Phantom" in name and ship_id == 35:
                 ship_id = 37
            #Change the names to match how they appear in Stations.jsonl
            if name == "Eagle":
                name = "Eagle Mk. II"
            if name == "Sidewinder":
                name = "Sidewinder Mk. I"
            if name == "Viper":
                name = "Viper Mk. III"

            # Make sure all the 'Mark N' ship names abbreviate 'Mark' the same.
            # Fix capitalization.
            name = name.replace('MK', 'Mk').replace('mk','Mk').replace('mK','Mk')
            # Fix no '.' in abbreviation.
            if "Mk" in name and "Mk." not in name:
                name = name.replace('Mk', 'Mk.')
            # Fix no trailing space.
            if "Mk." in name and "Mk. " not in name:
                name = name.replace("Mk.", "Mk. ")
            # Fix no leading space.
            if "Mk." in name and " Mk." not in name:
                name = name.replace("Mk.", " Mk.")
            
            tdenv.DEBUG1("Updating: {}, {}, {}, {}", ship_id, name, cost, fdev_id)
            try:
                self.execute("""INSERT INTO Ship
                                ( ship_id,name,cost,fdev_id ) VALUES
                                ( ?, ?, ?, ? ) """,
                                (ship_id, name, cost, fdev_id))
            except sqlite3.IntegrityError:
                try:
                    self.execute("""UPDATE Ship
                                    SET name = ?,cost = ?,fdev_id = ?
                                    WHERE ship_id = ?""",
                                    (name, cost, fdev_id,
                                     ship_id))
                except sqlite3.IntegrityError:
                    tdenv.DEBUG0("Unable to insert or update: {}, {}, {}, {}", ship_id, name, cost, fdev_id)

        self.updated['Ship'] = True
        
        tdenv.NOTE("Finished processing Ships. End time = {}", datetime.datetime.now())

    def importSystems(self):
        """
        Populate the System table using systems_populated.jsonl
        Writes directly to database.
        """
        tdb, tdenv = self.tdb, self.tdenv
        
        tdenv.NOTE("Processing Systems: Start time = {}", datetime.datetime.now())

        total = 1
        def blocks(f, size = 65536):
            while True:
                b = f.read(size)
                if not b: break
                yield b

        with open(str(self.dataPath / self.systemsPath), "r",encoding = "utf-8",errors = 'ignore') as f:
            total += (sum(bl.count("\n") for bl in blocks(f)))

        with open(str(self.dataPath / self.systemsPath), "rU") as fh:
            prog = pbar.Progress(total, 50)
            for line in fh:
                prog.increment(1, postfix=lambda value, goal: " " + str(round(value / total * 100)) + "%")
                system = json.loads(line)
                system_id = system['id']
                name = system['name']
                pos_x = system['x']
                pos_y = system['y']
                pos_z = system['z']
                modified = datetime.datetime.utcfromtimestamp(system['updated_at']).strftime('%Y-%m-%d %H:%M:%S')
                
                result = self.execute("SELECT modified FROM System WHERE system_id = ?", (system_id,)).fetchone()
                if result:
                    updated = timegm(datetime.datetime.strptime(result[0],'%Y-%m-%d %H:%M:%S').timetuple())
                    if system['updated_at'] > updated:
                        tdenv.DEBUG0("System '{}' has been updated: '{}' vs '{}'", name, modified, result[0])
                        tdenv.DEBUG1("Updating: {}, {}, {}, {}, {}, {}", system_id, name, pos_x, pos_y, pos_z, modified)
                        self.execute("""UPDATE System
                                    SET name = ?,pos_x = ?,pos_y = ?,pos_z = ?,modified = ?
                                    WHERE system_id = ?""", 
                                    (name, pos_x, pos_y, pos_z, modified,
                                     system_id))
                        self.updated['System'] = True
                else:
                    tdenv.DEBUG0("System '{}' has been added.", name)
                    tdenv.DEBUG1("Inserting: {}, {}, {}, {}, {}, {}", system_id, name, pos_x, pos_y, pos_z, modified)
                    self.execute("""INSERT INTO System
                                ( system_id,name,pos_x,pos_y,pos_z,modified ) VALUES
                                ( ?, ?, ?, ?, ?, ? ) """,
                                (system_id, name, pos_x, pos_y, pos_z, modified))
                    self.updated['System'] = True
            while prog.value < prog.maxValue:
                prog.increment(1, postfix=lambda value, goal: " " + str(round(value / total * 100)) + "%")
            prog.clear()
        
        tdenv.NOTE("Finished processing Systems. End time = {}", datetime.datetime.now())

    def importStations(self):
        """
        Populate the Station table using stations.jsonl
        Also populates the ShipVendor table if the option is set.
        Writes directly to database.
        """
        tdb, tdenv = self.tdb, self.tdenv
        
        tdenv.NOTE("Processing Stations, this may take a bit: Start time = {}", datetime.datetime.now())
        if self.getOption('shipvend'):
            tdenv.NOTE("Simultaneously processing ShipVendors.")

        if self.getOption('upvend'):
            tdenv.NOTE("Simultaneously processing UpgradeVendors, this will take quite a while.")


        total = 1
        def blocks(f, size = 65536):
            while True:
                b = f.read(size)
                if not b: break
                yield b

        with open(str(self.dataPath / self.stationsPath), "r",encoding = "utf-8",errors = 'ignore') as f:
            total += (sum(bl.count("\n") for bl in blocks(f)))
        
        with open(str(self.dataPath / self.stationsPath), "rU") as fh:
            prog = pbar.Progress(total, 50)
            for line in fh:
                prog.increment(1, postfix=lambda value, goal: " " + str(round(value / total * 100)) + "%")
                station = json.loads(line)
                
                # Import Stations
                station_id = station['id']
                name = station['name']
                system_id = station['system_id']
                ls_from_star = station['distance_to_star'] if station['distance_to_star'] else 0
                blackmarket = 'Y' if station['has_blackmarket'] else 'N'
                max_pad_size = station['max_landing_pad_size'] if station['max_landing_pad_size'] and station['max_landing_pad_size'] != 'None' else '?'
                market = 'Y' if station['has_market'] else 'N'
                shipyard = 'Y' if station['has_shipyard'] else 'N'
                modified = datetime.datetime.utcfromtimestamp(station['updated_at']).strftime('%Y-%m-%d %H:%M:%S')
                outfitting = 'Y' if station['has_outfitting'] else 'N'
                rearm = 'Y' if station['has_rearm'] else 'N'
                refuel = 'Y' if station['has_refuel'] else 'N'
                repair = 'Y' if station['has_repair'] else 'N'
                planetary = 'Y' if station['is_planetary'] else 'N'
                type_id = station['type_id'] if station['type_id'] else 0
                
                system = self.execute("SELECT System.name FROM System WHERE System.system_id = ?", (system_id,)).fetchone()[0].upper()
                
                result = self.execute("SELECT modified FROM Station WHERE station_id = ?", (station_id,)).fetchone()
                if result:
                    updated = timegm(datetime.datetime.strptime(result[0],'%Y-%m-%d %H:%M:%S').timetuple())
                    if station['updated_at'] > updated:
                        tdenv.DEBUG0("{}/{} has been updated: {} vs {}", 
                                    system ,name, modified, result[0])
                        tdenv.DEBUG1("Updating: {}, {}, {}, {}, {}, {}, {},"
                                              " {}, {}, {}, {}, {}, {}, {}, {}",
                                    station_id, name, system_id, ls_from_star, blackmarket,
                                    max_pad_size, market, shipyard, modified, outfitting,
                                    rearm, refuel, repair, planetary, type_id)
                        self.execute("""UPDATE Station
                                    SET name = ?, system_id = ?, ls_from_star = ?, blackmarket = ?,
                                    max_pad_size = ?, market = ?, shipyard = ?, modified = ?,
                                    outfitting = ?, rearm = ?, refuel = ?, repair = ?, planetary = ?, type_id = ?
                                    WHERE station_id = ?""", 
                                    (name, system_id, ls_from_star, blackmarket,
                                     max_pad_size, market, shipyard, modified,
                                     outfitting, rearm, refuel, repair, planetary, type_id,
                                     station_id))
                        self.updated['Station'] = True
                else:
                    tdenv.DEBUG0("{}/{} has been added:", system ,name)
                    tdenv.DEBUG1("Inserting: {}, {}, {}, {}, {}, {}, {},"
                                              " {}, {}, {}, {}, {}, {}, {}, {}",
                        station_id, name, system_id, ls_from_star, blackmarket,
                        max_pad_size, market, shipyard, modified, outfitting,
                        rearm, refuel, repair, planetary, type_id)
                    self.execute("""INSERT INTO Station (
                                station_id,name,system_id,ls_from_star,
                                blackmarket,max_pad_size,market,shipyard,
                                modified,outfitting,rearm,refuel,
                                repair,planetary,type_id ) VALUES
                                ( ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ? ) """,
                                (station_id,name,system_id,ls_from_star,
                                 blackmarket,max_pad_size,market,shipyard,
                                 modified,outfitting,rearm,refuel,
                                 repair,planetary,type_id))
                    self.updated['Station'] = True
                
                #Import shipyards into ShipVendors if shipvend is set.
                if station['has_shipyard'] and self.getOption('shipvend'):
                    if not station['shipyard_updated_at']:
                        station['shipyard_updated_at'] = station['updated_at']
                    modified = datetime.datetime.utcfromtimestamp(station['shipyard_updated_at']).strftime('%Y-%m-%d %H:%M:%S')
                    result = self.execute("SELECT modified FROM ShipVendor WHERE station_id = ?", (station_id,)).fetchone()
                    if result:
                        updated = timegm(datetime.datetime.strptime(result[0],'%Y-%m-%d %H:%M:%S').timetuple())
                    else:
                        updated = 0
                    if station['shipyard_updated_at'] > updated:
                        self.execute("DELETE FROM ShipVendor WHERE station_id = ?", (station_id,))
                        tdenv.DEBUG1("{}/{} has shipyard, updating ships sold.", system, name)
                        for ship in station['selling_ships']:
                            # Make sure all the 'Mark N' ship names abbreviate 'Mark' as '<Name> Mk. <Number>'.
                            # Fix capitalization.
                            ship = ship.replace('MK', 'Mk').replace('mk','Mk').replace('mK','Mk')
                            # Fix no '.' in abbreviation.
                            if "Mk" in ship and "Mk." not in ship:
                                ship = ship.replace('Mk', 'Mk.')
                            # Fix no trailing space.
                            if "Mk." in ship and "Mk. " not in ship:
                                ship = ship.replace("Mk.", "Mk. ")
                            # Fix no leading space.
                            if "Mk." in ship and " Mk." not in ship:
                                ship = ship.replace("Mk.", " Mk.")
                            
                            tdenv.DEBUG2("ship_id:{},station_id:{},modified:{}",
                                 ship,
                                 station_id,
                                 modified)
                            try:
                                self.execute("""INSERT INTO ShipVendor
                                        ( ship_id,station_id,modified ) VALUES
                                        ( (SELECT Ship.ship_id FROM Ship WHERE Ship.name = ?), ?, ? ) """,
                                        (ship,
                                         station_id,
                                         modified))
                            except sqlite3.IntegrityError:
                                continue
                        self.updated['ShipVendor'] = True
                        
                #Import Outfitters into UpgradeVendors if upvend is set.
                if station['has_outfitting'] and self.getOption('upvend'):
                    if not station['outfitting_updated_at']:
                        station['outfitting_updated_at'] = station['updated_at']
                    modified = datetime.datetime.utcfromtimestamp(station['outfitting_updated_at']).strftime('%Y-%m-%d %H:%M:%S')
                    result = self.execute("SELECT modified FROM UpgradeVendor WHERE station_id = ?", (station_id,)).fetchone()
                    if result:
                        updated = timegm(datetime.datetime.strptime(result[0],'%Y-%m-%d %H:%M:%S').timetuple())
                    else:
                        updated = 0
                    if station['outfitting_updated_at'] > updated:
                        self.execute("DELETE FROM UpgradeVendor WHERE station_id = ?", (station_id,))
                        tdenv.DEBUG1("{}/{} has outfitting, updating modules sold.", system, name)
                        for upgrade in station['selling_modules']:
                            tdenv.DEBUG2("upgrade_id:{},station_id:{},modified:{}",
                                 upgrade,
                                 station['id'],
                                 modified)
                            try:
                                self.execute("""INSERT INTO UpgradeVendor
                                        ( upgrade_id,station_id,cost,modified ) VALUES
                                        ( ?, ?, (SELECT Upgrade.cost FROM Upgrade WHERE Upgrade.upgrade_id = ?), ? ) """,
                                        (upgrade,
                                         station_id,
                                         upgrade,
                                         modified))
                            except sqlite3.IntegrityError:
                                continue
                        self.updated['UpgradeVendor'] = True
            while prog.value < prog.maxValue:
                prog.increment(1, postfix=lambda value, goal: " " + str(round(value / total * 100)) + "%")
            prog.clear()

        tdenv.NOTE("Finished processing Stations. End time = {}", datetime.datetime.now())

    def importCommodities(self):
        """
        Populate the Category, and Item tables using commodities.json
        Writes directly to the database.
        """
        tdb, tdenv = self.tdb, self.tdenv
                
        tdenv.NOTE("Processing Categories and Items: Start time = {}", datetime.datetime.now())
        with open(str(self.dataPath / self.commoditiesPath), "rU") as fh:
            commodities = json.load(fh)
            
        # EDDB still hasn't added these Commodities to the API,
        # so we'll add them ourselves.
        tdenv.NOTE("Checking for missing items....")

        # Need to get the category_ids from the .csv file.
        cat_ids = dict()
        try:
            with open(str(tdb.dataPath / Path("Category.csv")), "r") as fh:
                cats = csv.DictReader(fh, quotechar="'")
                for cat in cats:
                    cat_ids[cat['name']] =  int(cat['unq:category_id'])
        # Use default if no file, such as on a 'clean' run.
        except FileNotFoundError:
            cat_ids = {'Chemicals':1, 'Consumer Items':2, 'Legal Drugs':3, 'Foods':4, 'Industrial Materials':5,
                       'Machinery':6, 'Medicines':7, 'Metals':8, 'Minerals':9, 'Slavery':10, 'Technology':11,
                       'Textiles':12, 'Waste':13, 'Weapons':14, 'Unknown':15, 'Salvage':16}
        
        # EDCD is really quick about getting new items updated, so we'll use its item list to check
        # for missing items in EDDB.io's list.
        edcd_source = 'https://raw.githubusercontent.com/EDCD/FDevIDs/master/commodity.csv'
        edcd_csv = request.urlopen(edcd_source)
        edcd_dict = csv.DictReader(codecs.iterdecode(edcd_csv, 'utf-8'))
        
        def blank_item(name,ed_id,category,category_id):
            return {"id":ed_id,"name":name,"category_id":category_id,"average_price":None,"is_rare":0,"max_buy_price":None,"max_sell_price":None,"min_buy_price":None,"min_sell_price":None,"buy_price_lower_average":0,"sell_price_upper_average":0,"is_non_marketable":0,"ed_id":ed_id,"category":{"id":category_id,"name":category}}
        
        for line in iter(edcd_dict):
            if not any(c.get('ed_id', None) == int(line['id']) for c in commodities):
                tdenv.DEBUG0("'{}' with fdev_id {} not found, adding.", line['name'], line['id'])
                commodities.append(blank_item(line['name'],line['id'],line['category'],cat_ids[line['category']]))
        
        tdenv.NOTE("Missing item check complete.")
        
        # Prep-work for checking if an item's item_id has changed.
        cur_ids = dict()
        result = self.execute("SELECT fdev_id,item_id FROM Item ORDER BY fdev_id").fetchall()
        for item in result:
            cur_ids[item[0]] = item[1]
        
        tdenv.DEBUG0("Beginning loop.")
        for commodity in iter(commodities):
            # Make sure the broken item(s) in EDDB.io's API isn't imported. 
            if not commodity['ed_id']:
                tdenv.DEBUG0("Skipping faulty item: {}:{}" , commodity['id'], commodity['name'])
                continue
            # Get the categories from the json and place them into the Category table.
            category_id = commodity['category']['id']
            category_name = commodity['category']['name']
            
            tdenv.DEBUG1("Updating: {}, {}", category_id, category_name)
            try:
                self.execute("""INSERT INTO Category
                            ( category_id, name ) VALUES
                            ( ?, ? ) """,
                            (category_id, category_name))
            except sqlite3.IntegrityError:
                try:
                    self.execute("""UPDATE Category
                                SET name = ?
                                WHERE category_id = ?""", 
                                (category_name, category_id))

                except sqlite3.IntegrityError:
                    tdenv.DEBUG0("Unable to insert or update: {}, {}", category_id, category_name)
            
            # Only put regular items here, rare items can't be dealt with.
            if not commodity['is_rare']:
                item_id = commodity['id']
                name = commodity['name']
                if name.lower() == 'salvageable wreckage':
                    name = 'Wreckage Components'
                if name.lower() == 'political prisoner':
                    name = 'Political Prisoners'
                if name.lower() == 'hostage':
                    name = 'Hostages'
                if name.lower() == 'methanol monohydrate':
                    name = 'Methanol Monohydrate Crystals'
                if name.lower() == 'occupied cryopod':
                    name = 'Occupied Escape Pod'
                category_id = commodity['category_id']
                avg_price = commodity['average_price']
                fdev_id = commodity['ed_id']
                # "ui_order" doesn't have an equivalent field in the json.
                
                tdenv.DEBUG1("Updating: {}, {}, {}, {}, {}", item_id,name,category_id,avg_price,fdev_id)
                
                # If the item_id has changed, we need to completely delete the old entry.
                if cur_ids.get(fdev_id) != item_id:
                    tdenv.DEBUG1("Did not match item_id:{} with fdev_id:{} -- {}", item_id, fdev_id, cur_ids.get(fdev_id))
                    if cur_ids.get(fdev_id):
                        tdenv.DEBUG0("item_id  for '{}' has changed, updating.", name)
                        self.execute("DELETE FROM Item where fdev_id = ?", (fdev_id,))
                
                try:
                    self.execute("""INSERT INTO Item
                                ( item_id,name,category_id,avg_price,fdev_id ) VALUES
                                ( ?, ?, ?, ?, ? )""",
                                (item_id,name,category_id,avg_price,fdev_id))
                except sqlite3.IntegrityError:
                    try:
                        self.execute("""UPDATE Item
                                    SET name = ?,category_id = ?,avg_price = ?,fdev_id = ?
                                    WHERE item_id = ?""", 
                                    (name,category_id,avg_price,fdev_id,
                                     item_id))
                    except sqlite3.IntegrityError:
                        tdenv.DEBUG0("Unable to insert or update: {}, {}, {}, {}, {}", item_id,name,category_id,avg_price,fdev_id)
                        
        # The items aren't in the same order in the json as they are in the game's UI.
        # This creates a temporary object that has all the items sorted first
        # by category and second by name, as in the UI, which will then be used to
        # update the entries in the database with the correct "ui_order" value.
        temp = self.execute("""SELECT
                        name, category_id, fdev_id
                        FROM Item
                        ORDER BY category_id, name
                       """)
        cat_id = 0
        ui_order = 1
        tdenv.DEBUG0("Adding ui_order data to items.")
        for line in temp:
            if line[1] != cat_id:
                ui_order = 1
                cat_id = line[1]
            else:
                ui_order+= 1
            self.execute("""UPDATE Item
                        set ui_order = ?
                        WHERE fdev_id = ?""",
                       (ui_order, line[2]))
        
        self.updated['Category'] = True
        self.updated['Item'] = True
        
        tdenv.NOTE("Finished processing Categories and Items. End time = {}", datetime.datetime.now())

    def regenerate(self):
            for table in [
                "Category",
                "Item",
                "RareItem",
                "Ship",
                "ShipVendor",
                "Station",
                "System",
                "Upgrade",
                "UpgradeVendor",
            ]:
                if self.updated[table]:
                    _, path = csvexport.exportTableToFile(
                        self.tdb, self.tdenv, table
                    )
                    self.tdenv.NOTE("{} exported.", path)
    
    def commit(self):
        success = False
        while not success:
            try:
                self.tdb.getDB().commit()
                success = True
            except sqlite3.OperationalError:
                print("(commit) Database is locked, waiting for access.", end = "\r")
                time.sleep(1)

    def importListings(self, listings_file):
        """
        Updates the market data (AKA the StationItem table) using listings.csv
        Writes directly to database.
        """
        tdb, tdenv = self.tdb, self.tdenv
        
        tdenv.NOTE("Processing market data from {}: Start time = {}", listings_file, datetime.datetime.now())
        if not (self.dataPath / listings_file).exists():
            tdenv.NOTE("File not found, aborting: {}", (self.dataPath / listings_file))
            return
        
        total = 1
        
        from_live = 0 if listings_file == self.listingsPath else 1
        
        # Used to check if the listings file is using the fdev_id as a temporary
        # item_id, but the item is in the DB with a permanent item_id.
        fdev2item = dict()
        result = self.execute("SELECT fdev_id,item_id FROM Item ORDER BY fdev_id").fetchall()
        for item in result:
            fdev2item[item[0]] = item[1]
            
        
        def blocks(f, size = 65536):
            while True:
                b = f.read(size)
                if not b: break
                yield b
        
        with open(str(self.dataPath / listings_file), "r",encoding = "utf-8",errors = 'ignore') as f:
            total += (sum(bl.count("\n") for bl in blocks(f)))
        
        liveList = []
        liveStmt = """UPDATE StationItem
                    SET from_live = 0
                    WHERE station_id = ?"""
        
        delList = []
        delStmt = "DELETE from StationItem WHERE station_id = ?"
        
        listingList = []
        listingStmt = """INSERT OR IGNORE INTO StationItem
                        (station_id, item_id, modified,
                         demand_price, demand_units, demand_level,
                         supply_price, supply_units, supply_level, from_live)
                        VALUES ( ?, ?, ?, ?, ?, ?, ?, ?, ?, ? )"""
        
        items = []
        it_result = self.execute("SELECT item_id FROM Item ORDER BY item_id").fetchall()
        for item in it_result:
            items.append(item[0])
        
        with open(str(self.dataPath / listings_file), "rU") as fh:
            prog = pbar.Progress(total, 50)
            listings = csv.DictReader(fh)
            
            cur_station = -1
            
            for listing in listings:
                prog.increment(1, postfix=lambda value, goal: " " + str(round(value / total * 100)) + "%")
                
                station_id = int(listing['station_id'])
                
                if station_id != cur_station:
                    cur_station = station_id
                    skipStation = False
                    
                    # Check if listing already exists in DB and needs updated.
                    # Only need to check the date for the first item at a specific station.
                    result = self.execute("SELECT modified FROM StationItem WHERE station_id = ?", (station_id,)).fetchone()
                    if result:
                        updated = timegm(datetime.datetime.strptime(result[0],'%Y-%m-%d %H:%M:%S').timetuple())
                        # When the listings.csv data matches the database, update to make from_live == 0.
                        if int(listing['collected_at']) == updated and not from_live:
                            liveList.append((cur_station,))
                        # Unless the import file data is newer, nothing else needs to be done for this station,
                        # so the rest of the listings for this station can be skipped. 
                        if int(listing['collected_at']) <= updated:
                            skipStation = True
                            continue
                        
                        # The data from the import file is newer, so we need to delete the old data for this station.
                        delList.append((cur_station,))
                
                if skipStation:
                    continue
                
                # Since this station is not being skipped, get the data and prepare for insertion into the DB.
                item_id = int(listing['commodity_id'])
                # listings.csv includes rare items, which we are ignoring.
                if item_id not in items:
                    continue
                modified = datetime.datetime.utcfromtimestamp(int(listing['collected_at'])).strftime('%Y-%m-%d %H:%M:%S')
                demand_price = int(listing['sell_price'])
                demand_units = int(listing['demand'])
                demand_level = int(listing['demand_bracket']) if listing['demand_bracket'] != '' else -1
                supply_price = int(listing['buy_price'])
                supply_units = int(listing['supply'])
                supply_level = int(listing['supply_bracket']) if listing['supply_bracket'] != '' else -1
                
                listingList.append((station_id, item_id, modified,
                                    demand_price, demand_units, demand_level,
                                    supply_price, supply_units, supply_level, from_live))
            
            while prog.value < prog.maxValue:
                prog.increment(1, postfix=lambda value, goal: " " + str(round(value / total * 100)) + "%")
            prog.clear()
            
            tdenv.NOTE("Import file processing complete, updating database. {}", datetime.datetime.now())
            if liveList:
                tdenv.NOTE("Marking data now in the EDDB listings.csv as no longer 'live'. {}", datetime.datetime.now())
                self.executemany(liveStmt, liveList)
            if delList:
                tdenv.NOTE("Deleting old listing data. {}", datetime.datetime.now())
                self.executemany(delStmt, delList)
            if listingList:
                tdenv.NOTE("Inserting new listing data. {}", datetime.datetime.now())
                self.executemany(listingStmt, listingList)
        
        self.updated['Listings'] = True
        tdenv.NOTE("Finished processing market data. End time = {}", datetime.datetime.now())

    def run(self):
        tdb, tdenv = self.tdb, self.tdenv
        
        if self.getOption("progbar"):
            tdenv.NOTE("The 'progbar' option has been deprecated and no longer has any function.")
        
        #Create the /eddb folder for downloading the source files if it doesn't exist.
        try:
           Path(str(self.dataPath)).mkdir()
        except FileExistsError:
            pass
        
        # Run 'listings' by default:
        # If no options, or if only 'progbar', 'force', 'skipvend', and/or 'fallback', 
        # have been passed, enable 'listings'.
        default = True
        for option in self.options:
            if not ((option == 'force') or (option == 'fallback') or (option == 'skipvend') or (option == 'progbar')):
                default = False
        if default:
            self.options["listings"] = True
        
        # We can probably safely assume that the plugin has never been run if
        # the prices file doesn't exist, since the plugin always generates it.
        if not (tdb.dataPath / Path("TradeDangerous.prices")).exists():
            self.options["clean"] = True
        
        if self.getOption("clean"):
            # Rebuild the tables from scratch. Must be done on first run of plugin.
            # Can be done at anytime with the "clean" option.
            for name in [
                "Category",
                "Item",
                "Ship",
                "ShipVendor",
                "Station",
                "System",
                "Upgrade",
                "UpgradeVendor",
            ]:
                file = tdb.dataPath / Path(name + ".csv")
                try:
                    os.remove(str(file))
                except FileNotFoundError:
                    pass
                
            try:
                os.remove(str(tdb.dataPath) + "/TradeDangerous.db")
            except FileNotFoundError:
                pass
            try:
                os.remove(str(tdb.dataPath) + "/TradeDangerous.prices")
            except FileNotFoundError:
                pass

            # Because this is a clean run, we need to temporarily rename the RareItem.csv,
            # otherwise TD will crash trying to insert the rare items to the database,
            # because there's nothing in the Station table it tries to pull from.
            ri_path = tdb.dataPath / Path("RareItem.csv")
            rib_path = ri_path.with_suffix(".tmp")
            if rib_path.exists() and ri_path.exists():
                rib_path.unlink()
            if ri_path.exists():
                ri_path.rename(rib_path)

            tdb.reloadCache()
            
            # Now it's safe to move RareItems back.
            if ri_path.exists():
                ri_path.unlink()
            if rib_path.exists():
                rib_path.rename(ri_path)
                        
            self.options["all"] = True
            self.options["force"] = True

        tdenv.ignoreUnknown = True
        
        success = False
        while not success:
            try:
                tdb.load(maxSystemLinkLy = tdenv.maxSystemLinkLy)
                success = True
            except sqlite3.OperationalError:
                print("Database is locked, waiting for access.", end = "\r")
                time.sleep(1)
        
        #Select which options will be updated
        if self.getOption("listings"):
            self.options["item"] = True
            self.options["station"] = True

        if self.getOption("shipvend"):
            self.options["ship"] = True
            self.options["station"] = True

        if self.getOption("upvend"):
            self.options["upgrade"] = True
            self.options["station"] = True

        if self.getOption("station"):
            self.options["system"] = True

        if self.getOption("all"):
            self.options["item"] = True
            self.options["ship"] = True
            self.options["shipvend"] = True
            self.options["station"] = True
            self.options["system"] = True
            self.options["upgrade"] = True
            self.options["upvend"] = True
            self.options["listings"] = True

        if self.getOption("solo"):
            self.options["listings"] = False
            self.options["skipvend"] = True
        
        if self.getOption("skipvend"):
            self.options["shipvend"] = False
            self.options["upvend"] = False
        
        # Download required files and update tables.
        if self.getOption("upgrade"):
            if self.downloadFile(UPGRADES, self.upgradesPath) or self.getOption("force"):
                self.importUpgrades()
                self.commit()

        if self.getOption("ship"):
            if self.downloadFile(SHIPS_URL, self.shipsPath) or self.getOption("force"):
                self.importShips()
                self.commit()

        if self.getOption("system"):
            if self.downloadFile(SYSTEMS, self.systemsPath) or self.getOption("force"):
                self.importSystems()
                self.commit()

        if self.getOption("station"):
            if self.downloadFile(STATIONS, self.stationsPath) or self.getOption("force"):
                self.importStations()
                self.commit()

        if self.getOption("item"):
            if self.downloadFile(COMMODITIES, self.commoditiesPath) or self.getOption("force"):
                self.importCommodities()
                self.commit()
        
        #Remake the .csv files with the updated info.
        self.regenerate()

        if self.getOption("listings"):
            if self.downloadFile(LISTINGS, self.listingsPath) or self.getOption("force"):
                self.importListings(self.listingsPath)
            if not self.getOption("fallback") and (self.downloadFile(LIVE_LISTINGS, self.liveListingsPath) or self.getOption("force")):
                self.importListings(self.liveListingsPath)
        
        self.commit()
        
        tdb.close()
        
        if self.updated['Listings']:
            tdenv.NOTE("Regenerating .prices file.")
            cache.regeneratePricesFile(tdb, tdenv)

        tdenv.NOTE("Import completed.")

        # TD doesn't need to do anything, tell it to just quit.
        return False



#! /usr/bin/env python
# Small tool for comparing two system.csv files.
# Usage:
#  misc/compare-system-csvs.py <old file> <new file>
# Checks for stars that changed names, stars that moved
# and conflicts where two stars occupy the same space.

# Writes output suitable for "corrections.py" to stderr,
# so you could do something like:
#  $ diff-system-csvs.py oldSystem.csv newSystem.csv 2>corrections.txt

import sys

if len(sys.argv) != 3:
    raise SystemExit("Usage: {} <old file> <new file>".format(
                sys.argv[0]
            ))

import csv
import re
from pathlib import Path
from collections import namedtuple

class Loc(namedtuple('Loc', [ 'x', 'y', 'z' ])):
    def __str__(self):
        return "{},{},{}".format(self.x, self.y, self.z)


class Item(namedtuple('Item', [ 'norm', 'name', 'loc' ])):
    pass

normalizeRe = re.compile('[^A-Za-z0-9\' ]')

def readFile(filename):
    path = Path(filename)
    if not path.exists():
        raise SystemExit("File not found: {}".format(filename))

    names, locs = dict(), dict()

    with path.open("rU", encoding="utf-8") as fh:
        csvin = csv.reader(fh, delimiter=',', quotechar='\'', doublequote=True)
        # skip headings
        next(csvin)

        for line in csvin:
            name = line[0]
            x = float(line[1])
            y = float(line[2])
            z = float(line[3])

            normalized = normalizeRe.sub('', name).upper()
            try:
                prevEntry = names[normalized]
            except KeyError:
                pass
            else:
                print("Name clash: {}, this entry: {}, prev entry: {}".format(
                            normalized,
                            name,
                            prevEntry.name
                        ))
            item = Item(normalized, name, Loc(x, y, z))
            names[normalized] = item
            if item.loc in locs:
                print("{}: Duplicate location: {} and {} at {}".format(
                            filename, locs[item.loc].name, name, item.loc
                        ))
            else:
                locs[item.loc] = item

    return names, locs

oldNames, oldLocs = readFile(sys.argv[1])
newNames, newLocs = readFile(sys.argv[2])

for oldName, oldItem in oldNames.items():
    try:
        # Look the item up in the new names dict
        newItem = newNames[oldItem.norm]
    except:
        pass
    else:
        if oldItem.name != newItem.name:
            if oldItem.name.upper() == oldItem.name.upper():
                # Case changed, we can live with this.
                print("CAUTION: {} changed to {}".format(
                            oldItem.name,
                            newItem.name,
                        ))
            else:
                # Punctuation or something else, we need
                # a correction.
                print("{} became {}".format(
                            oldItem.name, newItem.name
                        ))
                print("  \"{}\": \"{}\",".format(
                            oldItem.name.upper(),
                            newItem.name,
                        ), file=sys.stderr)

        # Name didn't change, did the position?
        if oldItem.loc != newItem.loc:
            print("{} moved from {} -> {}".format(
                        oldItem.name, oldItem.loc, newItem.loc
                    ))

        # We don't need to do a location check on this one.
        try:
            del newLocs[newItem.loc]
        except KeyError:
            pass
        continue

    # We didn't find the old name in the new list, check
    # to see if there is a new star at the old position.
    try:
        newItem = newLocs[oldItem.loc]
    except:
        pass
    else:
        # we found something at the exact loc, which I
        # assume means the name changed.
        try:
            # we know it's there, we can remove it
            del newLocs[oldItem.loc]
        except KeyError:
            pass
        print("{} ({}) changed name to {}".format(
                    oldItem.name,
                    oldItem.loc,
                    newItem.name
                ))
        print("  \"{}\": \"{}\",".format(
                    oldItem.name.upper(),
                    newItem.name,
                ), file=sys.stderr)
        continue

    # we didn't find it, so as best we can tell it
    # has been removed. there's no easy way for us
    # to catch the case of a move and a minor reloc.
    print("{} ({}) was removed".format(
                oldItem.name,
                oldItem.loc
            ))
    print("  \"{}\": DELETED,".format(
                oldItem.name.upper(),
            ), file=sys.stderr)


for newLoc, newItem in newLocs.items():
    if newLoc in oldLocs:
        continue
    print("{} ({}) was added".format(newItem.name, newLoc))




from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from .commandenv import ResultRow
from .exceptions import CommandLineError
from .parsing import *
from ..formatting import RowFormat, ColumnFormat, max_len
from itertools import chain
from ..tradedb import AmbiguityError
from ..tradedb import System, Station
from ..tradedb import TradeDB

# Original by Dirk Wilhelm

from .. import csvexport
import re
import sys

######################################################################
# Parser config

help='List, add or update available ships to a station'
name='shipvendor'
epilog=None
arguments = [
    ParseArgument(
        'origin',
        help='Specify the full name of the station '
            '(SYS NAME/STN NAME is also supported).',
        metavar='STATIONNAME',
        type=str,
    ),
    ParseArgument(
        'ship',
        help='Comma or space separated list of ship names.',
        nargs='*',
    ),
]
switches = [
    MutuallyExclusiveGroup(
        ParseArgument(
            '--remove', '-rm',
            help='Indicates you want to remove an entry.',
            action='store_true',
        ),
        ParseArgument(
            '--add', '-a',
            help='Indicates you want to add a new station.',
            action='store_true',
        ),
        ParseArgument(
            '--name-sort',
            help='Sort listed ships by name.',
            action='store_true',
            dest='nameSort',
        ),
    ),
    ParseArgument(
        '--no-export',
        help='Do not update the .csv files.',
        action='store_true',
        dest='noExport',
    ),
]

######################################################################
# Helpers


def addShipVendor(tdb, cmdenv, station, ship):
    db = tdb.getDB()
    db.execute("""
                INSERT INTO ShipVendor (
                    ship_id, station_id
                ) VALUES (
                    ?, ?
                )
        """, [
                ship.ID, station.ID
        ])
    db.commit()
    cmdenv.NOTE("At {} adding {}", station.name(), ship.name())
    return ship


def removeShipVendor(tdb, cmdenv, station, ship):
    db = tdb.getDB()
    db.execute("""
            DELETE FROM ShipVendor WHERE ship_id = ? and station_id = ?
    """, [ship.ID,station.ID])
    db.commit()
    cmdenv.NOTE("At {} removing {}", station.name(), ship.name())
    return ship


def maybeExportToCSV(tdb, cmdenv):
    if cmdenv.noExport:
        cmdenv.DEBUG0("no-export set, not exporting stations")
        return

    lines, csvPath = csvexport.exportTableToFile(tdb, cmdenv, "ShipVendor")
    cmdenv.NOTE("{} updated.", csvPath)


def checkShipPresent(tdb, station, ship):
    # Ask the database how many rows it sees claiming
    # this ship is sold at that station. The value will
    # be zero if we don't have an entry, otherwise it
    # will be some positive number.
    cur = tdb.query("""
        SELECT  COUNT(*)
          FROM  ShipVendor
         WHERE  ship_id = ? AND station_id = ?
    """, [ship.ID, station.ID]
    )
    # Get the first result of the row and get the first column of that row
    count = int(cur.fetchone()[0])

    # Test if count > 0, if so, we'll return True, otherwise False
    return (count > 0)


def listShipsPresent(tdb, cmdenv, station, results):
    """ Populate results with a list of ships present at the given station """
    cur = tdb.query("""
        SELECT ship_id
          FROM ShipVendor
         WHERE station_id = ?
    """, [station.ID]
    )

    results.summary = ResultRow()
    results.summary.station = station
    ships = tdb.shipByID
    addShip = results.rows.append

    for (ship_id,) in cur:
        ship = ships.get(ship_id, None)
        if ship:
            addShip(ResultRow(ship=ship))

    if cmdenv.nameSort:
        results.rows.sort(key=lambda row: row.ship.name())
    else:
        results.rows.sort(key=lambda row: row.ship.cost, reverse=True)

    return results


######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    station = cmdenv.startStation
    if not isinstance(station, Station):
        raise CommandLineError(
            "{} is a system, not a station"
            .format(station.name())
        )
    if station.shipyard == 'N' and not cmdenv.remove:
        raise CommandLineError(
            "{} is flagged as having no shipyard."
            .format(station.name())
        )

    if cmdenv.add:
        action = addShipVendor
    elif cmdenv.remove:
        action = removeShipVendor
    else:
        return listShipsPresent(tdb, cmdenv, station, results)

    if not cmdenv.ship:
        raise CommandLineError(
            "No ship names specified."
        )

    cmdenv.NOTE("Using local database ({})", tdb.dbPath)
    ships = {}
    shipNames = chain.from_iterable(
        name.split(",") for name in cmdenv.ship
    )
    for shipName in shipNames:
        try:
            ship = tdb.lookupShip(shipName)
        except LookupError:
            raise CommandLineError("Unrecognized Ship: {}".format(shipName))

        # Lets see if that ship sails from the specified port.
        shipPresent = checkShipPresent(tdb, station, ship)
        if cmdenv.add:
            if shipPresent:
                cmdenv.DEBUG0("{} is already listed at {}",
                    ship.name(), station.name()
                )
            else:
                ships[ship.ID] = ship
        else:
            if not shipPresent:
                cmdenv.DEBUG0("{} is not listed at {}",
                    ship.name(), station.name()
                )
            else:
                ships[ship.ID] = ship

    if len(ships) == 0:
        cmdenv.NOTE("Nothing to do.")
        return None

    # We've checked that everything should be good.
    dataToExport = False
    for ship in ships.values():
        if action(tdb, cmdenv,station, ship):
            dataToExport = True

    maybeExportToCSV(tdb, cmdenv)

    return None

######################################################################
# Transform result set into output

def render(results, cmdenv, tdb):
    if not results or not results.rows:
        raise CommandLineError(
            "No ships available at {}"
            .format(results.summary.station.name())
        )

    maxShipLen = max_len(results.rows, key=lambda row: row.ship.name())

    rowFmt = RowFormat().append(
        ColumnFormat("Ship", '<', maxShipLen,
            key=lambda row: row.ship.name())
    ).append(
        ColumnFormat("Cost", '>', 12, 'n',
            key=lambda row: row.ship.cost)
    )

    if not cmdenv.quiet:
        heading, underline = rowFmt.heading()
        print(heading, underline, sep='\n')

    for row in results.rows:
        print(rowFmt.format(row))


"""This module should handle filesystem related operations
"""
from shutil import copy as shcopy
from os import makedirs, path, utime
from pathlib import Path

__all__ = ['copy', 'copyallfiles', 'touch', 'ensurefolder']

def pathify(*args):
    if len(args) > 1 or not isinstance(args[0], Path):
        return Path(*args)
    return args[0]

def copy(src, dst):
    """
    copy src to dst
    takes string or Path object as input
    returns Path(dst) on success
    raises FileNotFoundError if src does not exist
    """
    srcPath = pathify(src).resolve()
    dstPath = pathify(dst)
    shcopy(str(srcPath), str(dstPath))
    return dstPath

def copy_if_newer(src, dst):
    """
    copy src to dst if src is newer 
    takes string or Path object as input
    returns Path(dst) on success
    returns Path(src) if not newer
    raises FileNotFoundError if src does not exist
    """
    srcPath = pathify(src).resolve()
    dstPath = pathify(dst)
    if dstPath.exists() and not (dstPath.stat().st_mtime < srcPath.stat().st_mtime):
        return srcPath
    else:
        shcopy(str(srcPath), str(dstPath))
        return dstPath
    
def copyallfiles(srcdir, dstdir):
    """
    Copies all files in srcdir to dstdir
    """
    srcPath = pathify(srcdir)
    dstPath = pathify(dstdir)

    for p in srcPath.glob('*.*'):
        if p.is_file():
            copy(p, dstPath / p.name)

def touch(filename):
    """
    Creates file if it doesn't exist.
    Always modifies utime.
    Returns a Path(filename)
    """
    path = pathify(filename)
    path.touch(exist_ok=True)
    return path

def ensureflag(flagfile, action=None):
    """Checks if flagfile exist and IF NOT the action function
    will be executed. The flagfile will be 'touched' at the end
    
    Parameters
    ----------
    flagfile : string
        path to the file used as flag
    action : callable
        this will be called if the flagfile doesn't exist
    
    Returns
    -------
    Path(flagfile)
    """
    flagPath = pathify(flagfile)
    if not flagPath.exists() and callable(action):
        action()
    return touch(flagPath)

def ensurefolder(folder):
    """Creates the folder if it doesn't exist
    
    Parameters
    ----------
    folder : string|pathlib.Path
        path to the folder
    
    Returns
    -------
    pathlib.Path
        Resolved path of the folder
    """
    folderPath = pathify(folder)
    try:
        makedirs(str(folderPath))
    except FileExistsError:
        pass
    return folderPath.resolve()



# --------------------------------------------------------------------
# Copyright (C) Oliver 'kfsone' Smith 2014 <oliver@kfs.org>:
# Copyright (C) Bernd 'Gazelle' Gollesch 2016, 2017
# Copyright (C) Jonathan 'eyeonus' Jones 2018
#
# You are free to use, redistribute, or even print and eat a copy of
# this software so long as you include this copyright notice.
# I guarantee there is at least one bug neither of us knew about.
# --------------------------------------------------------------------
# TradeDangerous :: Modules :: Generate TradeDangerous.prices

from __future__ import absolute_import, with_statement, print_function, division, unicode_literals

import sys
import os
import re
import sqlite3


class Element(object):
    basic     = (1 << 0)
    supply    = (1 << 1)
    timestamp = (1 << 2)
    full      = (basic | supply | timestamp)
    blanks    = (1 <<31)


######################################################################
# Main

def dumpPrices(
            dbPath,             # Path() or str
            elementMask,        # which columns to output
            stationID=None,     # limits to one station
            file=None,          # file handle to write to
            defaultZero=False,
            debug=0
    ):
    """
        Generate a prices list using data from the DB.
        If stationID is not none, only the specified station is dumped.
        If file is not none, outputs to the given file handle.
    """
    
    withTimes  = (elementMask & Element.timestamp)
    getBlanks  = (elementMask & Element.blanks)
    
    conn = sqlite3.connect(str(dbPath))
    conn.execute("PRAGMA foreign_keys=ON")
    cur  = conn.cursor()
    
    systems = { ID: name for (ID, name) in cur.execute("SELECT system_id, name FROM System") }
    stations = {
        ID: [ name, systems[sysID] ]
        for (ID, name, sysID)
        in cur.execute("SELECT station_id, name, system_id FROM Station")
    }
    categories = { ID: name for (ID, name) in cur.execute("SELECT category_id, name FROM Category") }
    items = {
        ID: [ name, catID, categories[catID] ]
        for (ID, name, catID)
        in cur.execute("SELECT item_id, name, category_id FROM Item")
    }
    
    # find longest item name
    longestName = max(items.values(), key=lambda ent: len(ent[0]))
    longestNameLen = len(longestName[0])
    
    if stationID:
        # check if there are prices for the station
        cur.execute("""
            SELECT  COUNT(*)
              FROM  StationItem
             WHERE station_id = {}
        """.format(stationID))
        if not cur.fetchone()[0]:
            getBlanks = True
    
    defaultDemandVal = 0 if defaultZero else -1
    if stationID:
        stationWhere = "WHERE stn.station_id = {}".format(stationID)
    else:
        stationWhere = ""
    
    if getBlanks:
        itemJoin = "LEFT OUTER"
    else:
        itemJoin = "INNER"
    
    cur.execute("SELECT CURRENT_TIMESTAMP")
    now = cur.fetchone()[0]
    
    stmt = """
        SELECT  stn.station_id, itm.item_id
                , IFNULL(si.demand_price, 0)
                , IFNULL(si.supply_price, 0)
                , IFNULL(si.demand_units, {defDemand})
                , IFNULL(si.demand_level, {defDemand})
                , IFNULL(si.supply_units, {defDemand})
                , IFNULL(si.supply_level, {defDemand})
                , si.modified
          FROM  Station stn,
                Category AS cat
                INNER JOIN Item AS itm USING (category_id)
                {itemJoin} JOIN StationItem AS si
                    ON (si.station_id = stn.station_id
                        AND si.item_id = itm.item_id)
                {stationWhere}
         ORDER  BY stn.station_id, cat.name, itm.ui_order
    """
    
    sql = stmt.format(
        stationWhere=stationWhere,
        defDemand=defaultDemandVal,
        itemJoin=itemJoin,
    )
    if debug:
        print(sql)
    cur.execute(sql)
    
    lastSys, lastStn, lastCat = None, None, None
    
    if not file: file = sys.stdout
    
    if stationID:
        stationSet = str(stations[stationID])
    else:
        stationSet = "ALL Systems/Stations"
    
    file.write(
        "# TradeDangerous prices for {}\n"
        "\n"
        "# REMOVE ITEMS THAT DON'T APPEAR IN THE UI\n"
        "# ORDER IS REMEMBERED: Move items around within categories "
            "to match the game UI\n"
        "\n"
        "# File syntax:\n"
        "# <item name> <sell> <buy> [<demand> <supply> [<timestamp>]]\n"
        "#   Use '?' for demand/supply when you don't know/care,\n"
        "#   Use '-' for demand/supply to indicate unavailable,\n"
        "#   Otherwise use a number followed by L, M or H, e.g.\n"
        "#     1L, 23M or 30000H\n"
        "# If you omit the timestamp, the current time will be used when "
            "the file is loaded.\n"
        "\n".format(
            stationSet
    ))
    
    levelDesc = "?0LMH"
    maxCrWidth = 7
    levelWidth = 9
    
    outFmt = (
        "      {{:<{width}}}"
        " {{:>{crwidth}}}"
        " {{:>{crwidth}}}"
        "  {{:>{lvlwidth}}}"
        " {{:>{lvlwidth}}}".format(
            width=longestNameLen,
            crwidth=maxCrWidth,
            lvlwidth=levelWidth,
        )
    )
    if withTimes:
        outFmt += "  {}"
    outFmt += "\n"
    output = outFmt.format(
        "Item Name",
        "SellCr", "BuyCr",
        "Demand", "Supply",
        "Timestamp",
    )
    file.write('#' + output[1:])
    
    naIQL = "-"
    unkIQL = "?"
    defIQL = "?" if not defaultZero else "-"
    
    output = ""
    for (stnID, itemID, fromStn, toStn, demand, demandLevel, supply, supplyLevel, modified) in cur:
        modified = modified or now
        station, system = stations[stnID]
        item, catID, category = items[itemID]
        if stnID != lastStn:
            file.write(output)
            output = "\n\n@ {}/{}\n".format(system.upper(), station)
            lastStn = stnID
            lastCat = None

        if catID is not lastCat:
            output += "   + {}\n".format(category)
            lastCat = catID
        
        # Is this item on sale?
        if toStn > 0:
            # Zero demand-price gets default demand, which will
            # be either unknown or zero depending on -0.
            # If there is a price, always default to unknown
            # because it can be sold here but the demand is just
            # not useful as data.
            demandStr = defIQL if fromStn <= 0 else unkIQL
            if supplyLevel == 0:
                supplyStr = naIQL
            elif supplyLevel < 0 and supply <= 0:
                supplyStr = defIQL
            else:
                units = "?" if supply < 0 else str(supply)
                level = levelDesc[supplyLevel + 1]
                supplyStr = units + level
        else:
            if fromStn == 0 or demandLevel == 0:
                demandStr = naIQL
            elif demandLevel < 0 and demand <= 0:
                demandStr = defIQL
            else:
                units = "?" if demand < 0 else str(demand)
                level = levelDesc[demandLevel + 1]
                demandStr = units + level
            supplyStr = naIQL
        output += outFmt.format(
                    item,
                    fromStn, toStn,
                    demandStr, supplyStr,
                    modified
                )
    
    file.write(output)

if __name__ == "__main__":
    import tradedb
    tdb = tradedb.TradeDB(load=False)
    dumpPrices(tdb.dbPath, elementMask=Element.full)



import csv
import pathlib
import random

from .. import cache, tradeenv, transfers, csvexport
from ..tradedb import Category, Item
from . import PluginException, ImportPluginBase

class ImportPlugin(ImportPluginBase):
    """
    Download and process EDCD CSV-file(s)
    """

    pluginOptions = {
        'local':      "Use local EDCD CSV-files.",
        'csvs':       "Download and process all EDCD CSV-files.",
        'shipyard':   "Download and process EDCD shipyard.csv",
        'commodity':  "Download and process EDCD commodity.csv",
        'outfitting': "Download and process EDCD outfitting.csv",
    }

    def __init__(self, tdb, tdenv):
        super().__init__(tdb, tdenv)

    def check_local_edcd(self):
        """
            Check local DB against EDCD
        """
        tdb, tdenv = self.tdb, self.tdenv

        tdCategories = list()
        for catID, catTD in sorted(tdb.categories(), key=lambda x: x[1].dbname):
            tdCategories.append(catTD.dbname)
            if catTD.dbname not in self.edcdCategories:
                tdenv.WARN("Category '{}' not in EDCD", catTD.dbname)

            for itemTD in sorted(catTD.items, key=lambda x: x.dbname):
                itemEDCD = self.edcdItems.get(itemTD.dbname, None)
                if not itemEDCD:
                    tdenv.DEBUG0("Item '{}' not in EDCD", itemTD.fullname)
                else:
                    if catTD.dbname != itemEDCD.category.dbname:
                        tdenv.WARN("Item '{}' has different category "
                            "'{}' (TD) != '{}' (EDCD)",
                            itemTD.dbname,
                            catTD.dbname, itemEDCD.category.dbname
                        )
        self.tdCategories = tdCategories

    def update_item_order(self, db):
        """
            Update the ui_order of the items
        """
        sqlStmt = (
            "SELECT category_id, name"
             " FROM Category"
            " ORDER BY name"
        )
        itmStmt = (
            "SELECT item_id, name"
             " FROM Item"
            " WHERE category_id = ?"
            " ORDER BY name"
        )
        updStmt = (
            "UPDATE Item"
              " SET ui_order = ?"
            " WHERE item_id = ?"
        )
        for catID, catName in db.execute(sqlStmt):
            itmOrder = 0
            for itmID, itmName in db.execute(itmStmt, [ catID ]):
                itmOrder += 1
                db.execute(updStmt, [ itmOrder, itmID ])

    def check_edcd_local(self):
        """
            Check EDCD items against local DB
        """
        tdb, tdenv = self.tdb, self.tdenv

        addItem = 0
        updItem = 0
        addCategory = 0

        commit = False
        db = tdb.getDB()
        for catNameEDCD in sorted(self.edcdCategories):
            catEDCD = self.edcdCategories[catNameEDCD]
            if catEDCD.dbname not in self.tdCategories:
                tdenv.NOTE("New category '{}'", catEDCD.dbname)
                sqlStmt = "INSERT INTO Category(name) VALUES(?)"

                tdenv.DEBUG0("SQL-Statement: {}", sqlStmt)
                tdenv.DEBUG0("SQL-Values: {}", [ catEDCD.dbname ])

                db.execute(sqlStmt, [ catEDCD.dbname ])
                addCategory += 1
                commit = True

            # Check EDCD items against local DB
            for itemEDCD in sorted(catEDCD.items, key=lambda x: x.dbname):
                itemTD = tdb.itemByName.get(itemEDCD.dbname, None)
                if not itemTD:
                    itemTD = tdb.itemByFDevID.get(itemEDCD.fdevID, None)
                    if itemTD:
                        tdenv.WARN(
                            "Item '{}' has different name '{}' (TD) != '{} '(EDCD).",
                            itemEDCD.fdevID, itemEDCD.dbname, itemTD.dbname
                        )
                    else:
                        tdenv.NOTE("New Item '{}'", itemEDCD.fullname)
                        insColumns = [
                            "name",
                            "category_id",
                            "fdev_id"
                        ]
                        insValues = [
                            itemEDCD.dbname,
                            catEDCD.dbname,
                            itemEDCD.fdevID
                        ]
                        if itemEDCD.avgPrice:
                            insColumns.append('avg_price')
                            insValues.append(itemEDCD.avgPrice)
                        sqlStmt = (
                            "INSERT INTO Item({}) VALUES(?,"
                                "(SELECT category_id "
                                   "FROM Category WHERE name = ?),?{})"
                            .format(
                                ",".join(insColumns),
                                ",?" if itemEDCD.avgPrice else ""
                            )
                        )
                        tdenv.DEBUG0("SQL-Statement: {}", sqlStmt)
                        tdenv.DEBUG0("SQL-Values: {}", insValues)

                        db.execute(sqlStmt, insValues)
                        addItem += 1
                        commit = True
                else:
                    updValues = list()
                    updColumns = list()
                    if itemEDCD.avgPrice and itemTD.avgPrice != itemEDCD.avgPrice:
                        updColumns.append('avg_price')
                        updValues.append(itemEDCD.avgPrice)
                    if not itemTD.fdevID:
                        updColumns.append('fdev_id')
                        updValues.append(itemEDCD.fdevID)
                    else:
                        if itemTD.fdevID != itemEDCD.fdevID:
                            tdenv.WARN(
                                "Item '{}' has different FDevID {} (TD) != {} (EDCD).",
                                itemTD.fullname, itemTD.fdevID, itemEDCD.fdevID
                            )
                    if len(updColumns):
                        tdenv.NOTE("Update Item '{}' {} {}",
                            itemTD.fullname, updColumns, updValues
                        )
                        updValues.append(itemTD.ID)
                        sqlStmt = (
                            "UPDATE Item SET {} = ?"
                            " WHERE Item.item_id = ?".format(
                            " = ?,".join(updColumns))
                        )
                        tdenv.DEBUG0("SQL-Statement: {}", sqlStmt)
                        tdenv.DEBUG0("SQL-Values: {}", updValues)

                        db.execute(sqlStmt, updValues)
                        updItem += 1
                        commit = True

        if addItem:
            self.update_item_order(db)

        if commit:
            db.commit()
            if addCategory:
                tdenv.NOTE("Added {} categorie(s)", addCategory)
                _, csvPath = csvexport.exportTableToFile(tdb, tdenv, 'Category')
                tdenv.NOTE("{} updated.", csvPath)
            if addItem or updItem:
                for textAdd, intAdd in [('Added', addItem), ('Updated', updItem)]:
                    if intAdd:
                        tdenv.NOTE("{} {} item(s)", textAdd, intAdd)
                _, csvPath = csvexport.exportTableToFile(tdb, tdenv, 'Item')
                tdenv.NOTE("{} updated.", csvPath)

        if not commit:
            tdenv.NOTE("Nothing had to be done")


    def process_fdevids_items(self, localPath, tableName):
        """
            More to do for commodities
        """
        tdb, tdenv = self.tdb, self.tdenv
        tdenv.NOTE("Processing {}", tableName)

        itmCount = 0
        catCount = 0
        edcdItems = {}
        edcdCategories = {}
        with localPath.open('rU', encoding='utf-8') as importFile:
            csvIn = csv.reader(
                importFile, delimiter=',', quotechar="'", doublequote=True
            )
            # first line must be the column names
            columnDefs = next(csvIn)
            columnCount = len(columnDefs)

            tdenv.DEBUG0("columnDefs: {}", columnDefs)

            idxMap = {}
            for cIndex, cName in enumerate(columnDefs):
                idxMap[cName] = cIndex
            tdenv.DEBUG0("idxMap: {}", idxMap)

            notGood = False
            for checkMe in ('id', 'category', 'name'):
                if checkMe not in idxMap:
                    tdenv.WARN("'{}' column not in {}.", checkMe, localPath)
                    notGood = True
            hasAverage = True if 'average' in idxMap else False

            def castToInteger(val):
                try:
                    val = int(val)
                except:
                    val = None
                    pass
                return val

            if notGood:
                tdenv.NOTE("Import stopped.", checkMe, localPath)
            else:
                for lineIn in csvIn:
                    if not lineIn: continue
                    lineNo = csvIn.line_num
                    tdenv.DEBUG0("LINE {}: {}", lineNo, lineIn)

                    # strip() it do be save
                    edcdId       = lineIn[idxMap['id']].strip()
                    edcdCategory = lineIn[idxMap['category']].strip()
                    edcdName     = lineIn[idxMap['name']].strip()
                    if hasAverage:
                        edcdAverage = lineIn[idxMap['average']].strip()
                    else:
                        edcdAverage = None

                    if edcdCategory.lower() == "nonmarketable":
                        # we are, after all, a trading tool
                        tdenv.DEBUG0("Ignoring {}/{}", edcdCategory, edcdName)
                        continue

                    edcdId      = castToInteger(edcdId)
                    edcdAverage = castToInteger(edcdAverage)

                    itemCategory = edcdCategories.get(edcdCategory, None)
                    if not itemCategory:
                        itemCategory = Category(0, edcdCategory, [])
                        edcdCategories[edcdCategory] = itemCategory
                        catCount += 1
                        tdenv.DEBUG1("NEW CATEGORY: {}", edcdCategory)

                    if edcdName not in edcdItems:
                        newItem = Item(
                            edcdId, edcdName, itemCategory,
                            '{}/{}'.format(itemCategory.dbname, edcdName),
                            edcdAverage, edcdId
                        )
                        edcdItems[edcdName] = newItem
                        itemCategory.items.append(newItem)
                        itmCount += 1
                        tdenv.DEBUG1("NEW ITEM: {}", edcdName)

        self.edcdItems = edcdItems
        self.edcdCategories = edcdCategories

        tdenv.NOTE("Found {} categorie(s)", catCount)
        tdenv.NOTE("Found {} item(s)", itmCount)

        if catCount or itmCount:
            self.check_local_edcd()
            self.check_edcd_local()

    def process_fdevids_table(self, localPath, tableName):
        """
            Shipyard and outfitting files can be directly imported.
        """
        tdb, tdenv = self.tdb, self.tdenv
        tdenv.NOTE("Processing {}", tableName)

        db = tdb.getDB()
        db.execute("DELETE FROM {}".format(tableName))
        cache.processImportFile(
            tdenv,
            db,
            localPath,
            tableName
        )
        db.commit()
        lines, csvPath = csvexport.exportTableToFile(
            tdb,
            tdenv,
            tableName,
        )
        tdenv.NOTE("Imported {} {}(s).", lines, tableName)
        tdenv.NOTE("{} updated.", csvPath)

    def download_fdevids(self):
        """
            Download the current data from EDCD,
            see https://github.com/EDCD/FDevIDs
        """
        tdb, tdenv = self.tdb, self.tdenv

        BASE_URL = "https://raw.githubusercontent.com/EDCD/FDevIDs/master/"
        downloadList = []
        if self.getOption("shipyard"):
            downloadList.append(
                ('FDevShipyard', 'shipyard.csv', self.process_fdevids_table)
            )
        if self.getOption("outfitting"):
            downloadList.append(
                ('FDevOutfitting', 'outfitting.csv', self.process_fdevids_table)
            )
        if self.getOption("commodity"):
            downloadList.append(
                ('FDevCommodity', 'commodity.csv', self.process_fdevids_items)
            )

        if len(downloadList) == 0:
            tdenv.NOTE("I don't know what do to, give me some options!")
            return

        optLocal = self.getOption("local")
        for tableName, fileEDCD, callMeBack in downloadList:
            localPath = tdb.dataPath / pathlib.Path(tableName).with_suffix(".csv")
            if optLocal:
                if not localPath.exists():
                    raise PluginException(
                        "CSV-file '{}' not found.".format(str(localPath))
                    )
            else:
                transfers.download(
                    tdenv,
                    BASE_URL + fileEDCD,
                    str(localPath),
                )
            if callMeBack:
                callMeBack(localPath, tableName)

    def run(self):
        tdb, tdenv = self.tdb, self.tdenv

        tdenv.DEBUG0("csvs: {}", self.getOption("csvs"))
        tdenv.DEBUG0("shipyard: {}", self.getOption("shipyard"))
        tdenv.DEBUG0("commodity: {}", self.getOption("commodity"))
        tdenv.DEBUG0("outfitting: {}", self.getOption("outfitting"))

        if self.getOption("csvs"):
            self.options["shipyard"] = True
            self.options["commodity"] = True
            self.options["outfitting"] = True

        # Ensure the cache is built and reloaded.
        tdb.reloadCache()
        tdb.load(maxSystemLinkLy=tdenv.maxSystemLinkLy)

        self.download_fdevids()

        # We did all the work
        return False
